"""archetypal ZoneLoad."""

import collections
import logging as lg
import math
import sqlite3
from enum import Enum

import numpy as np
import pandas as pd
from sigfig import round
from validator_collection import checkers, validators

from archetypal import settings
from archetypal.template.schedule import UmiSchedule
from archetypal.template.umi_base import UmiBase
from archetypal.utils import log, reduce, timeit


class DimmingTypes(Enum):
    """DimmingType class."""

    Continuous = 0
    Off = 1
    Stepped = 2

    def __lt__(self, other):
        """Assert if self is lower then other."""
        return self._value_ < other._value_

    def __gt__(self, other):
        """Assert if self is greater then other."""
        return self._value_ > other._value_


class ZoneLoad(UmiBase):
    """Zone Loads.

    Important:
        Please note that the calculation of the equipment power density will sum
        up the electric equipment objects as well as the gas equipment objects.

    .. image:: ../images/template/zoneinfo-loads.png
    """

    __slots__ = (
        "_dimming_type",
        "_equipment_availability_schedule",
        "_lights_availability_schedule",
        "_occupancy_schedule",
        "_equipment_power_density",
        "_illuminance_target",
        "_lighting_power_density",
        "_people_density",
        "_is_equipment_on",
        "_is_lighting_on",
        "_is_people_on",
        "_area",
        "_volume",
    )

    def __init__(
        self,
        Name,
        EquipmentPowerDensity=0,
        EquipmentAvailabilitySchedule=None,
        LightingPowerDensity=0,
        LightsAvailabilitySchedule=None,
        PeopleDensity=0,
        OccupancySchedule=None,
        IsEquipmentOn=True,
        IsLightingOn=True,
        IsPeopleOn=True,
        DimmingType=DimmingTypes.Continuous,
        IlluminanceTarget=500,
        area=1,
        volume=1,
        **kwargs,
    ):
        """Initialize a new ZoneLoad object.

        Args:
            DimmingType (int): Different types to dim the lighting to respect the
                IlluminanceTarget and taking into account the daylight illuminance:
                    - Continuous = 0, the overhead lights dim continuously and
                      linearly from (maximum electric power, maximum light output) to (
                      minimum electric power, minimum light output) as the daylight
                      illuminance increases. The lights stay on at the minimum point
                      with further increase in the daylight illuminance.
                    - Off = 1, Lights switch off completely when the minimum
                      dimming point is reached.
                    - Stepped = 2, the electric power input and light output vary
                      in discrete, equally spaced steps.
            EquipmentAvailabilitySchedule (UmiSchedule): The name of
                the schedule (Day | Week | Year) that modifies the design level
                parameter for electric equipment.
            EquipmentPowerDensity (float): Equipment Power Density in the zone
                (W/m²).
            IlluminanceTarget (float): Number of lux to be respected in the zone
            LightingPowerDensity (float): Lighting Power Density in the zone
                (W/m²).
            LightsAvailabilitySchedule (UmiSchedule): The name of the
                schedule (Day | Week | Year) that modifies the design level
                parameter for lighting.
            OccupancySchedule (UmiSchedule): The name of the schedule
                (Day | Week | Year) that modifies the number of people parameter
                for electric equipment.
            IsEquipmentOn (bool): If True, heat gains from Equipment are taken
                into account for the zone's load calculation.
            IsLightingOn (bool): If True, heat gains from Lights are taken into
                account for the zone's load calculation.
            IsPeopleOn (bool): If True, heat gains from People are taken into
                account for the zone's load calculation.
            PeopleDensity (float): Density of people in the zone (people/m²).
            area (float): The floor area assiciated to this zone load object.
            **kwargs: Other keywords passed to the parent constructor :class:`UmiBase`.
        """
        super(ZoneLoad, self).__init__(Name, **kwargs)

        self.EquipmentPowerDensity = EquipmentPowerDensity
        self.EquipmentAvailabilitySchedule = EquipmentAvailabilitySchedule
        self.LightingPowerDensity = LightingPowerDensity
        self.LightsAvailabilitySchedule = LightsAvailabilitySchedule
        self.PeopleDensity = PeopleDensity
        self.OccupancySchedule = OccupancySchedule
        self.IsEquipmentOn = IsEquipmentOn
        self.IsLightingOn = IsLightingOn
        self.IsPeopleOn = IsPeopleOn
        self.DimmingType = DimmingTypes(DimmingType)
        self.IlluminanceTarget = IlluminanceTarget
        self.area = area
        self.volume = volume

    @property
    def DimmingType(self):
        """Get or set the dimming type.

        Hint:
            To set the value an int or a string is supported.
            Choices are (<DimmingTypes.Continuous: 0>, <DimmingTypes.Off: 1>,
            <DimmingTypes.Stepped: 2>)
        """
        return self._dimming_type

    @DimmingType.setter
    def DimmingType(self, value):
        if checkers.is_string(value):
            assert DimmingTypes[value], (
                f"Input value error for '{value}'. "
                f"Expected one of {tuple(a for a in DimmingTypes)}"
            )
            self._dimming_type = DimmingTypes[value]
        elif checkers.is_numeric(value):
            assert DimmingTypes[value], (
                f"Input value error for '{value}'. "
                f"Expected one of {tuple(a for a in DimmingTypes)}"
            )
            self._dimming_type = DimmingTypes(value)
        elif isinstance(value, DimmingTypes):
            self._dimming_type = value
        else:
            raise ValueError(f"Could not set DimmingType with value '{value}'")

    @property
    def EquipmentAvailabilitySchedule(self):
        """Get or set the equipment availability schedule."""
        return self._equipment_availability_schedule

    @EquipmentAvailabilitySchedule.setter
    def EquipmentAvailabilitySchedule(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input value error for '{value}'. Value must be of type '"
                f"{UmiSchedule}', not {type(value)}"
            )
            # set quantity on schedule as well
            value.quantity = self.EquipmentPowerDensity
        self._equipment_availability_schedule = value

    @property
    def EquipmentPowerDensity(self):
        """Get or set the equipment power density [W/m²]."""
        return self._equipment_power_density

    @EquipmentPowerDensity.setter
    def EquipmentPowerDensity(self, value):
        self._equipment_power_density = validators.float(
            value, minimum=0, allow_empty=True
        )

    @property
    def IlluminanceTarget(self):
        """Get or set the illuminance target [lux]."""
        return self._illuminance_target

    @IlluminanceTarget.setter
    def IlluminanceTarget(self, value):
        self._illuminance_target = validators.float(value, minimum=0)

    @property
    def LightingPowerDensity(self):
        """Get or set the lighting power density [W/m²]."""
        return self._lighting_power_density

    @LightingPowerDensity.setter
    def LightingPowerDensity(self, value):
        self._lighting_power_density = validators.float(
            value, minimum=0, allow_empty=True
        )

    @property
    def LightsAvailabilitySchedule(self) -> UmiSchedule:
        """Get or set the lights availability schedule."""
        return self._lights_availability_schedule

    @LightsAvailabilitySchedule.setter
    def LightsAvailabilitySchedule(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input value error for '{value}'. Value must be of type '"
                f"{UmiSchedule}', not {type(value)}"
            )
            # set quantity on schedule as well
            value.quantity = self.LightingPowerDensity
        self._lights_availability_schedule = value

    @property
    def OccupancySchedule(self) -> UmiSchedule:
        """Get or set the occupancy schedule."""
        return self._occupancy_schedule

    @OccupancySchedule.setter
    def OccupancySchedule(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input value error for '{value}'. Value must be if type '"
                f"{UmiSchedule}', not {type(value)}"
            )
            # set quantity on schedule as well
            value.quantity = self.PeopleDensity
        self._occupancy_schedule = value

    @property
    def PeopleDensity(self):
        """Get or set the people density [ppl/m²]."""
        return self._people_density

    @PeopleDensity.setter
    def PeopleDensity(self, value):
        self._people_density = validators.float(value, minimum=0)

    @property
    def IsEquipmentOn(self):
        """Get or set the use of equipment [bool]."""
        return self._is_equipment_on

    @IsEquipmentOn.setter
    def IsEquipmentOn(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsEquipmentOn must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_equipment_on = value

    @property
    def IsLightingOn(self):
        """Get or set the use of lighting [bool]."""
        return self._is_lighting_on

    @IsLightingOn.setter
    def IsLightingOn(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsLightingOn must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_lighting_on = value

    @property
    def IsPeopleOn(self):
        """Get or set people [bool]."""
        return self._is_people_on

    @IsPeopleOn.setter
    def IsPeopleOn(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsPeopleOn must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_people_on = value

    @property
    def area(self):
        """Get or set the floor area of the zone associated to this zone load [m²]."""
        return self._area

    @area.setter
    def area(self, value):
        self._area = validators.float(value, minimum=0)

    @property
    def volume(self):
        """Get or set the volume of the zone associated to this zone load [m³]."""
        return self._volume

    @volume.setter
    def volume(self, value):
        self._volume = validators.float(value, minimum=0)

    @classmethod
    def from_dict(cls, data, schedules, **kwargs):
        """Create a ZoneLoad from a dictionary.

        Args:
            data (dict): A python dictionary with the structure shown bellow.
            schedules (dict): A python dictionary of UmiSchedules with their id as keys.
            **kwargs: keywords passed to parent constructors.

        .. code-block:: python

            {
              "$id": "172",
              "DimmingType": 1,
              "EquipmentAvailabilitySchedule": {
                "$ref": "147"
              },
              "EquipmentPowerDensity": 8.0,
              "IlluminanceTarget": 500.0,
              "LightingPowerDensity": 12.0,
              "LightsAvailabilitySchedule": {
                "$ref": "146"
              },
              "OccupancySchedule": {
                "$ref": "145"
              },
              "IsEquipmentOn": true,
              "IsLightingOn": true,
              "IsPeopleOn": true,
              "PeopleDensity": 0.055,
              "Category": "Office Spaces",
              "Comments": null,
              "DataSource": "MIT_SDL",
              "Name": "B_Off_0 loads"
            },
        """
        _id = data.pop("$id")
        return cls(
            id=_id,
            EquipmentAvailabilitySchedule=schedules[
                data.pop("EquipmentAvailabilitySchedule")["$ref"]
            ],
            LightsAvailabilitySchedule=schedules[
                data.pop("LightsAvailabilitySchedule")["$ref"]
            ],
            OccupancySchedule=schedules[data.pop("OccupancySchedule")["$ref"]],
            **data,
            **kwargs,
        )

    @classmethod
    @timeit
    def from_zone(cls, zone, zone_ep, **kwargs):
        """Create a ZoneLoad object from a :class:`ZoneDefinition`.

        Args:
            zone_ep:
            zone (ZoneDefinition): zone to gets information from
            kwargs: keywords passed to the parent constructor.
        """
        # If Zone is not part of total area, it should not have a ZoneLoad object.
        if not zone._is_part_of_total_floor_area:
            return None

        # Get schedule index for different loads and create ZoneLoad arguments
        # Verify if Equipment in zone

        # create database connection with sqlite3
        with sqlite3.connect(str(zone_ep.theidf.sql_file)) as conn:
            sql_query = "select ifnull(ZoneIndex, null) from Zones where ZoneName=?"
            t = (zone.Name.upper(),)
            c = conn.cursor()
            c.execute(sql_query, t)
            (zone_index,) = c.fetchone()

            sql_query = "select t.* from NominalElectricEquipment t where ZoneIndex=?"
            nominal_elec = pd.read_sql(sql_query, conn, params=(zone_index,))

            sql_query = "select t.* from NominalGasEquipment t where ZoneIndex=?"
            nominal_gas = pd.read_sql(sql_query, conn, params=(zone_index,))

            def get_schedule(series):
                """Compute the schedule with quantity for nominal equipment series."""
                sched = series["ScheduleIndex"]
                sql_query = (
                    "select t.ScheduleName, t.ScheduleType as M from "
                    "Schedules t where ScheduleIndex=?"
                )
                sched_name, sched_type = c.execute(sql_query, (int(sched),)).fetchone()
                level_ = float(series["DesignLevel"])
                if level_ > 0:
                    return UmiSchedule.from_epbunch(
                        zone_ep.theidf.schedules_dict[sched_name.upper()],
                        quantity=level_,
                    )

            schedules = []
            if not nominal_elec.empty:
                # compute schedules series
                elec_scds = nominal_elec.apply(get_schedule, axis=1).to_list()
                elec_scds = list(filter(None, elec_scds))
                schedules.extend(elec_scds)

            if not nominal_gas.empty:
                # compute schedules series
                gas_scds = nominal_gas.apply(get_schedule, axis=1).to_list()
                gas_scds = list(filter(None, gas_scds))
                schedules.extend(gas_scds)

            if schedules:
                EquipmentAvailabilitySchedule = reduce(
                    UmiSchedule.combine,
                    schedules,
                    quantity=True,
                )
                EquipmentPowerDensity = (
                    EquipmentAvailabilitySchedule.quantity / zone.area
                )
            else:
                EquipmentAvailabilitySchedule = None
                EquipmentPowerDensity = np.NaN

            # Verifies if Lights in zone
            sql_query = "select t.* from NominalLighting t where ZoneIndex=?"
            nominal_lighting = pd.read_sql(sql_query, conn, params=(zone_index,))

            lighting_schedules = []
            if not nominal_lighting.empty:
                # compute schedules series
                light_scds = nominal_lighting.apply(get_schedule, axis=1)
                lighting_schedules.extend(light_scds)

            if lighting_schedules:
                LightsAvailabilitySchedule = reduce(
                    UmiSchedule.combine,
                    lighting_schedules,
                    quantity=True,
                )
                LightingPowerDensity = LightsAvailabilitySchedule.quantity / zone.area
            else:
                LightsAvailabilitySchedule = None
                LightingPowerDensity = np.NaN

            # Verifies if People in zone

            def get_schedule(series):
                """Compute schedule with quantity for nominal equipment series."""
                sched = series["NumberOfPeopleScheduleIndex"]
                sql_query = (
                    "select t.ScheduleName, t.ScheduleType as M from "
                    "Schedules t where ScheduleIndex=?"
                )
                sched_name, sched_type = c.execute(sql_query, (int(sched),)).fetchone()
                return UmiSchedule.from_epbunch(
                    zone_ep.theidf.schedules_dict[sched_name.upper()],
                    quantity=series["NumberOfPeople"],
                )

            sql_query = "select t.* from NominalPeople t where ZoneIndex=?"
            nominal_people = pd.read_sql(sql_query, conn, params=(zone_index,))

            occupancy_schedules = []
            if not nominal_people.empty:
                # compute schedules series
                occ_scds = nominal_people.apply(get_schedule, axis=1)
                occupancy_schedules.extend(occ_scds)

            if occupancy_schedules:
                OccupancySchedule = reduce(
                    UmiSchedule.combine,
                    occupancy_schedules,
                    quantity=lambda x: sum(obj.quantity for obj in x),
                )
                PeopleDensity = OccupancySchedule.quantity / zone.area
            else:
                OccupancySchedule = None
                PeopleDensity = np.NaN

        name = zone.Name + "_ZoneLoad"
        z_load = cls(
            Name=name,
            DimmingType=_resolve_dimming_type(zone, zone_ep),
            EquipmentAvailabilitySchedule=EquipmentAvailabilitySchedule,
            EquipmentPowerDensity=EquipmentPowerDensity,
            IlluminanceTarget=_resolve_illuminance_target(zone, zone_ep),
            LightingPowerDensity=LightingPowerDensity,
            LightsAvailabilitySchedule=LightsAvailabilitySchedule,
            OccupancySchedule=OccupancySchedule,
            IsEquipmentOn=EquipmentPowerDensity > 0,
            IsLightingOn=LightingPowerDensity > 0,
            IsPeopleOn=PeopleDensity > 0,
            PeopleDensity=PeopleDensity,
            Category=zone.DataSource,
            area=zone.area,
            volume=zone.volume,
            **kwargs,
        )
        return z_load

    def combine(self, other, weights=None):
        """Combine two ZoneLoad objects together. Returns a new object.

        Args:
            other (ZoneLoad): The other ZoneLoad object.
            weights (list-like, optional): A list-like object of len 2. If None,
                the `settings.zone_weight` of the objects is used.

        Returns:
            (ZoneLoad): the combined ZoneLoad object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        meta = self._get_predecessors_meta(other)

        if not weights:
            zone_weight = settings.zone_weight
            weights = [
                getattr(self, str(zone_weight)),
                getattr(other, str(zone_weight)),
            ]
            log(
                'using zone {} "{}" as weighting factor in "{}" '
                "combine.".format(
                    zone_weight,
                    " & ".join(list(map(str, map(int, weights)))),
                    self.__class__.__name__,
                )
            )

        new_attr = dict(
            DimmingType=max(self.DimmingType, other.DimmingType),
            EquipmentAvailabilitySchedule=UmiSchedule.combine(
                self.EquipmentAvailabilitySchedule,
                other.EquipmentAvailabilitySchedule,
                weights=[self.area, other.area],
                quantity=True,
            ),
            EquipmentPowerDensity=self.float_mean(
                other, "EquipmentPowerDensity", weights
            ),
            IlluminanceTarget=self.float_mean(other, "IlluminanceTarget", weights),
            LightingPowerDensity=self.float_mean(
                other, "LightingPowerDensity", weights
            ),
            LightsAvailabilitySchedule=UmiSchedule.combine(
                self.LightsAvailabilitySchedule,
                other.LightsAvailabilitySchedule,
                weights=[self.area, other.area],
                quantity=True,
            ),
            OccupancySchedule=UmiSchedule.combine(
                self.OccupancySchedule,
                other.OccupancySchedule,
                weights=[self.area, other.area],
                quantity=True,
            ),
            IsEquipmentOn=any([self.IsEquipmentOn, other.IsEquipmentOn]),
            IsLightingOn=any([self.IsLightingOn, other.IsLightingOn]),
            IsPeopleOn=any([self.IsPeopleOn, other.IsPeopleOn]),
            PeopleDensity=self.float_mean(other, "PeopleDensity", weights),
        )

        new_obj = self.__class__(
            **meta, **new_attr, allow_duplicates=self.allow_duplicates
        )
        new_obj.area = self.area + other.area
        new_obj.volume = self.volume + other.volume
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def validate(self):
        """Validate object and fill in missing values."""
        if not self.DimmingType:
            self.DimmingType = DimmingTypes.Continuous
        if not self.EquipmentAvailabilitySchedule:
            self.EquipmentAvailabilitySchedule = UmiSchedule.constant_schedule()
        if not self.EquipmentPowerDensity:
            self.EquipmentPowerDensity = 0
        if not self.IlluminanceTarget:
            self.IlluminanceTarget = 500
        if not self.LightingPowerDensity:
            self.LightingPowerDensity = 0
        if not self.LightsAvailabilitySchedule:
            self.LightsAvailabilitySchedule = UmiSchedule.constant_schedule()
        if not self.OccupancySchedule:
            self.OccupancySchedule = UmiSchedule.constant_schedule()
        if not self.IsEquipmentOn:
            self.IsEquipmentOn = False
        if not self.IsLightingOn:
            self.IsLightingOn = False
        if not self.IsPeopleOn:
            self.IsPeopleOn = False
        if not self.PeopleDensity:
            self.PeopleDensity = 0
        return self

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            DimmingType=self.DimmingType,
            EquipmentAvailabilitySchedule=self.EquipmentAvailabilitySchedule,
            EquipmentPowerDensity=self.EquipmentPowerDensity,
            IlluminanceTarget=self.IlluminanceTarget,
            LightingPowerDensity=self.LightingPowerDensity,
            LightsAvailabilitySchedule=self.LightsAvailabilitySchedule,
            OccupancySchedule=self.OccupancySchedule,
            IsEquipmentOn=self.IsEquipmentOn,
            IsLightingOn=self.IsLightingOn,
            IsPeopleOn=self.IsPeopleOn,
            PeopleDensity=self.PeopleDensity,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def to_dict(self):
        """Return ZoneLoad dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["DimmingType"] = self.DimmingType.value
        data_dict[
            "EquipmentAvailabilitySchedule"
        ] = self.EquipmentAvailabilitySchedule.to_ref()
        data_dict["EquipmentPowerDensity"] = (
            round(self.EquipmentPowerDensity, 3)
            if not math.isnan(self.EquipmentPowerDensity)
            else 0
        )
        data_dict["IlluminanceTarget"] = round(self.IlluminanceTarget, 3)
        data_dict["LightingPowerDensity"] = (
            round(self.LightingPowerDensity, 3)
            if not math.isnan(self.LightingPowerDensity)
            else 0
        )
        data_dict[
            "LightsAvailabilitySchedule"
        ] = self.LightsAvailabilitySchedule.to_ref()
        data_dict["OccupancySchedule"] = self.OccupancySchedule.to_ref()
        data_dict["IsEquipmentOn"] = self.IsEquipmentOn
        data_dict["IsLightingOn"] = self.IsLightingOn
        data_dict["IsPeopleOn"] = self.IsPeopleOn
        data_dict["PeopleDensity"] = (
            round(self.PeopleDensity, 3) if not math.isnan(self.PeopleDensity) else 0
        )
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def to_epbunch(self, idf, zone_name):
        """Convert the zone load to epbunch given an idf model and a zone name.

        Args:
            idf (IDF): The idf model. epbunches will be added to this model.
            zone_name (str): The name of the zone in the idf model.

        .. code-block:: python

            People,
                People Perim,             !- Name
                Perim,                    !- Zone or ZoneList Name
                B_Off_Y_Occ,              !- Number of People Schedule Name
                People/Area,              !- Number of People Calculation Method
                ,                         !- Number of People
                0.055,                    !- People per Zone Floor Area
                ,                         !- Zone Floor Area per Person
                0.3,                      !- Fraction Radiant
                AUTOCALCULATE,            !- Sensible Heat Fraction
                PerimPeopleActivity,      !- Activity Level Schedule Name
                3.82e-08,                 !- Carbon Dioxide Generation Rate
                No,                       !- Enable ASHRAE 55 Comfort Warnings
                ZoneAveraged,             !- Mean Radiant Temperature Calculation Type
                ,                         !- Surface NameAngle Factor List Name
                PerimWorkEfficiency,      !- Work Efficiency Schedule Name
                DynamicClothingModelASHRAE55,    !- Clothing Insulation Calculation Method
                ,                         !- Clothing Insulation Calculation Method Schedule Name
                ,                         !- Clothing Insulation Schedule Name
                PerimAirVelocity,         !- Air Velocity Schedule Name
                AdaptiveASH55;            !- Thermal Comfort Model 1 Type

            Lights,
                Perim General lighting,    !- Name
                Perim,                    !- Zone or ZoneList Name
                B_Off_Y_Lgt,              !- Schedule Name
                Watts/Area,               !- Design Level Calculation Method
                ,                         !- Lighting Level
                12,                       !- Watts per Zone Floor Area
                ,                         !- Watts per Person
                0,                        !- Return Air Fraction
                0.42,                     !- Fraction Radiant
                0.18,                     !- Fraction Visible
                1,                        !- Fraction Replaceable
                ;                         !- EndUse Subcategory

            ElectricEquipment,
                Perim Equipment 1,        !- Name
                Perim,                    !- Zone or ZoneList Name
                B_Off_Y_Plg,              !- Schedule Name
                Watts/Area,               !- Design Level Calculation Method
                ,                         !- Design Level
                8,                        !- Watts per Zone Floor Area
                ,                         !- Watts per Person
                0,                        !- Fraction Latent
                0.2,                      !- Fraction Radiant
                0,                        !- Fraction Lost
                ;                         !- EndUse Subcategory

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        people = idf.newidfobject(
            "PEOPLE",
            Name=self.Name,
            Zone_or_ZoneList_Name=zone_name,
            Number_of_People_Schedule_Name=self.OccupancySchedule.to_epbunch(idf).Name,
            Number_of_People_Calculation_Method="People/Area",
            People_per_Zone_Floor_Area=self.PeopleDensity,
            Fraction_Radiant=0.3,
            Sensible_Heat_Fraction="AUTOCALCULATE",
            Activity_Level_Schedule_Name=idf.newidfobject(
                "SCHEDULE:CONSTANT", Name="PeopleActivity", Hourly_Value=125.28
            ).Name,
            Carbon_Dioxide_Generation_Rate=3.82e-08,
            Enable_ASHRAE_55_Comfort_Warnings="No",
            Mean_Radiant_Temperature_Calculation_Type="ZoneAveraged",
            Work_Efficiency_Schedule_Name=idf.newidfobject(
                "SCHEDULE:CONSTANT", Name="WorkEfficiency", Hourly_Value=0
            ).Name,
            Clothing_Insulation_Calculation_Method="DynamicClothingModelASHRAE55",
            Air_Velocity_Schedule_Name=idf.newidfobject(
                "SCHEDULE:CONSTANT", Name="AirVelocity", Hourly_Value=0.2
            ).Name,
        )
        lights = idf.newidfobject(
            key="LIGHTS",
            Name=self.Name,
            Zone_or_ZoneList_Name=zone_name,
            Schedule_Name=self.LightsAvailabilitySchedule.to_epbunch(idf).Name,
            Design_Level_Calculation_Method="Watts/Area",
            Watts_per_Zone_Floor_Area=self.LightingPowerDensity,
            Return_Air_Fraction=0,
            Fraction_Radiant=0.42,
            Fraction_Visible=0.18,
            Fraction_Replaceable=1,
        )
        equipment = idf.newidfobject(
            "ELECTRICEQUIPMENT",
            Name=self.Name,
            Zone_or_ZoneList_Name=zone_name,
            Schedule_Name=self.EquipmentAvailabilitySchedule.to_epbunch(idf).Name,
            Design_Level_Calculation_Method="Watts/Area",
            Watts_per_Zone_Floor_Area=self.EquipmentPowerDensity,
            Fraction_Latent=0,
            Fraction_Radiant=0.2,
            Fraction_Lost=0,
        )
        return people, lights, equipment

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(
            **self.mapping(validate=False), area=self.area, volume=self.volume
        )

    def __add__(self, other):
        """Combine self and other."""
        return self.combine(other)

    def __hash__(self):
        """Return the hash value of self."""
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __key__(self):
        """Get a tuple of attributes. Useful for hashing and comparing."""
        return (
            self.DimmingType,
            self.EquipmentAvailabilitySchedule,
            self.EquipmentPowerDensity,
            self.IlluminanceTarget,
            self.LightingPowerDensity,
            self.LightsAvailabilitySchedule,
            self.OccupancySchedule,
            self.IsEquipmentOn,
            self.IsLightingOn,
            self.IsPeopleOn,
            self.PeopleDensity,
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, ZoneLoad):
            return NotImplemented
        else:
            return self.__key__() == other.__key__()


def _resolve_dimming_type(zone, zone_ep):
    """Resolve the dimming type for the Zone object.

    Args:
        zone_ep:
    """
    # First, retrieve the list of Daylighting objects for this zone. Uses the eppy
    # `getreferingobjs` method.
    possible_ctrls = zone_ep.getreferingobjs(
        iddgroups=["Daylighting"], fields=["Zone_Name"]
    )
    # Then, if there are controls
    if possible_ctrls:
        # Filter only the "Daylighting:Controls"
        ctrls = [
            ctrl
            for ctrl in possible_ctrls
            if ctrl.key.upper() == "Daylighting:Controls".upper()
        ]
        ctrl_types = [ctrl["Lighting_Control_Type"] for ctrl in ctrls]

        # There should only be one control per zone. A set of controls should return 1.
        if len(set(ctrl_types)) == 1:
            dimming_type, *_ = set(ctrl_types)
            if dimming_type.lower() not in ["continuous", "stepped"]:
                raise ValueError(
                    f"A dimming type of type '{dimming_type}' for zone '{zone.Name}' is not yet supported in UMI"
                )
            else:
                log(f"Dimming type for zone '{zone.Name}' set to '{dimming_type}'")
                return DimmingTypes[dimming_type]  # Return first element
        else:
            raise ValueError(
                "Could not resolve more than one dimming types for Zone {}. "
                "Make sure there is only one".format(zone.Name)
            )
    else:
        # Else, there are no dimming controls => set to "Off".
        log(
            "No dimming type found for zone {}. Setting as Off".format(zone.Name),
            lg.DEBUG,
        )
        return DimmingTypes.Off


def _resolve_illuminance_target(zone, zone_ep):
    """Resolve the illuminance target for the Zone object.

    Args:
        zone_ep:
    """
    # First, retrieve the list of Daylighting objects for this zone. Uses the eppy
    # `getreferingobjs` method.
    possible_ctrls = zone_ep.getreferingobjs(
        iddgroups=["Daylighting"], fields=["Zone_Name"]
    )
    # Then, if there are controls
    if possible_ctrls:
        # Filter only the "Daylighting:Controls"
        ctrls = [
            ctrl
            for ctrl in possible_ctrls
            if ctrl.key.upper() == "Daylighting:Controls".upper()
        ]
        ctrl_types = [
            ctrl["Illuminance_Setpoint_at_Reference_Point_1"] for ctrl in ctrls
        ]

        # There should only be one control per zone. A set of controls should return 1.
        if len(set(ctrl_types)) == 1:
            dimming_type = next(iter(set(ctrl_types)))
            log(f"Illuminance target for zone '{zone.Name}' set to '{dimming_type}'")
            return float(dimming_type)  # Return first element
        else:
            raise ValueError(
                "Could not resolve more than one illuminance targets for Zone {}. "
                "Make sure there is only one".format(zone.Name)
            )
    else:
        # Else, there are no dimming controls => set to "Off".
        log(
            "No illuminance target found for zone {}. Setting to default 500 "
            "lux".format(zone.Name),
            lg.DEBUG,
        )
        return 500



################################################################################
# Module: __init__.py
# Description: Archetypal: Retrieve, construct and analyse building archetypes
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

# Version of the package
from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution("archetypal").version
except DistributionNotFound:
    # package is not installed
    __version__ = "0.0.0"  # should happen only if package is copied, not installed.
else:
    # warn if a newer version of archetypal is available
    from outdated import warn_if_outdated
    from .eplus_interface.version import warn_if_not_compatible
finally:
    # warn if energyplus not installed or incompatible
    from .eplus_interface.version import warn_if_not_compatible

    warn_if_not_compatible()

# don't display futurewarnings

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)
warnings.simplefilter(action="ignore", category=UserWarning)

from .idfclass import IDF
from .umi_template import UmiTemplateLibrary



"""archetypal umi template library."""

__all__ = [
    "BuildingTemplate",
    "ZoneConditioning",
    "DomesticHotWaterSetting",
    "GasMaterial",
    "GlazingMaterial",
    "ZoneLoad",
    "OpaqueConstruction",
    "OpaqueMaterial",
    "UmiSchedule",
    "StructureInformation",
    "VentilationSetting",
    "WindowConstruction",
    "WindowSetting",
    "ZoneDefinition",
    "ZoneConstructionSet",
]

from archetypal.template.building_template import BuildingTemplate
from archetypal.template.conditioning import ZoneConditioning
from archetypal.template.constructions.opaque_construction import OpaqueConstruction
from archetypal.template.constructions.window_construction import WindowConstruction
from archetypal.template.dhw import DomesticHotWaterSetting
from archetypal.template.load import ZoneLoad
from archetypal.template.materials import GasMaterial, GlazingMaterial, OpaqueMaterial
from archetypal.template.schedule import UmiSchedule
from archetypal.template.structure import StructureInformation
from archetypal.template.ventilation import VentilationSetting
from archetypal.template.window_setting import WindowSetting
from archetypal.template.zone_construction_set import ZoneConstructionSet
from archetypal.template.zonedefinition import ZoneDefinition



import glob
import sys

import pytest

# Parametrization of the fixture scratch_then_cache. The following array
# tells pytest to use True than False for all tests that use this fixture.
# This is very usefull to test the behavior of methods that use cached data
# or not.
from archetypal import settings, utils

do = [True, False]


@pytest.fixture(params=do, ids=["from_scratch", "from_cache"], scope="function")
def scratch_then_cache(request):
    """# remove the tests/temp folder if it already exists so we
    start fresh with tests"""
    # request is a special parameter known to pytest. It passes whatever is in
    # params=do. Ids are there to give the test a human readable name.
    if request.param:
        dirs = [
            settings.data_folder,
            settings.cache_folder,
            settings.imgs_folder,
        ]
        for dir in dirs:
            dir.rmtree_p()


samples_ = ["regular", "umi_samples"]  # ['problematic', 'regular', 'umi_samples']


@pytest.fixture(params=samples_, ids=samples_, scope="session")
def idf_source(request):
    return glob.glob("tests/input_data/{}/*.idf".format(request.param))


@pytest.fixture(scope="session")
def config():
    utils.config(
        data_folder="tests/.temp/data",
        logs_folder="tests/.temp/logs",
        imgs_folder="tests/.temp/images",
        cache_folder="tests/.temp/cache",
        use_cache=True,
        log_file=False,
        log_console=True,
        umitemplate="tests/input_data/umi_samples/BostonTemplateLibrary_2.json",
        debug=True,
    )


@pytest.fixture(scope="class")
def clean_config(config):
    """calls config fixture and clears default folders"""

    dirs = [settings.data_folder, settings.cache_folder, settings.imgs_folder]
    for dir in dirs:
        dir.rmtree_p()


# List fixtures that are located outiside of conftest.py so that they can be
# used in other tests
pytest_plugins = ["tests.test_dataportals"]

ALL = set("darwin linux win32".split())


def pytest_runtest_setup(item):
    supported_platforms = ALL.intersection(mark.name for mark in item.iter_markers())
    plat = sys.platform
    if supported_platforms and plat not in supported_platforms:
        pytest.skip("cannot run on platform %s" % (plat))


# dynamically define files to be ignored
collect_ignore = ["test_core.py"]


def get_platform():
    """Returns the MacOS release number as tuple of ints"""
    import platform

    release, versioninfo, machine = platform.mac_ver()
    release_split = release.split(".")
    return tuple(map(safe_int_cast, release_split))


def safe_int_cast(val, default=0):
    """Safely casts a value to an int"""
    try:
        return int(val)
    except (ValueError, TypeError):
        return default



"""archetypal OpaqueConstruction."""

import collections
import uuid

import numpy as np
from eppy.bunch_subclass import BadEPFieldError
from validator_collection import validators

from archetypal.template.constructions.base_construction import LayeredConstruction
from archetypal.template.materials.material_layer import MaterialLayer
from archetypal.template.materials.opaque_material import OpaqueMaterial


class OpaqueConstruction(LayeredConstruction):
    """Opaque Constructions.

    .. image:: ../images/template/constructions-opaque.png

    Properties:
        * r_value
        * u_value
        * r_factor
        * u_factor
        * equivalent_heat_capacity_per_unit_volume
        * specific_heat
        * heat_capacity_per_unit_wall_area
        * total_thickness
        * mass_per_unit_area
        * timeconstant_per_unit_area
        * solar_reflectance_index
    """

    __slots__ = ("area",)

    def __init__(self, Name, Layers, **kwargs):
        """Initialize an OpaqueConstruction.

        Args:
            Layers (list of archetypal.MaterialLayer): List of MaterialLayers making
                up the construction.
            **kwargs: Other attributes passed to parent constructors such as
                :class:`ConstructionBase`.
        """
        super(OpaqueConstruction, self).__init__(Name, Layers, **kwargs)
        self.area = 1

    @property
    def r_value(self):
        """Get or set the thermal resistance [K⋅m2/W] (excluding air films).

        Note that, when setting the R-value, the thickness of the inferred
        insulation layer will be adjusted.
        """
        return super(OpaqueConstruction, self).r_value

    @r_value.setter
    def r_value(self, value):
        # First, find the insulation layer
        i = self.infer_insulation_layer()
        all_layers_except_insulation_layer = [a for a in self.Layers]
        all_layers_except_insulation_layer.pop(i)
        insulation_layer: MaterialLayer = self.Layers[i]

        if value <= sum([a.r_value for a in all_layers_except_insulation_layer]):
            raise ValueError(
                f"Cannot set assembly r-value smaller than "
                f"{sum([a.r_value for a in all_layers_except_insulation_layer])} "
                f"because it would result in an insulation of a "
                f"negative thickness. Try a higher value or changing the material "
                f"layers instead."
            )

        alpha = float(value) / self.r_value
        new_r_value = (
            ((alpha - 1) * sum([a.r_value for a in all_layers_except_insulation_layer]))
        ) + alpha * insulation_layer.r_value
        insulation_layer.r_value = new_r_value

    @property
    def equivalent_heat_capacity_per_unit_volume(self):
        """Get the equivalent per unit wall volume heat capacity [J/(kg⋅K)].

        Hint:
            "The physical quantity which represents the heat storage capability
            is the wall heat capacity, defined as HC=M·c. While the per unit
            wall area of this quantity is (HC/A)=ρ·c·δ, where δ the wall
            thickness, the per unit volume wall heat capacity, being a
            characteristic wall quantity independent from the wall thickness, is
            equal to ρ·c. This quantity for a composite wall of an overall
            thickness L, is usually defined as the equivalent per unit wall
            volume heat capacity and it is expressed as
            :math:`{{(ρ·c)}}_{eq}{{=(1/L)·∑}}_{i=1}^n{{(ρ}}_i{{·c}}_i{{·δ}}_i{)}`
            where :math:`{ρ}_i`, :math:`{c}_i` and :math:`{δ}_i` are the
            densities, the specific heat capacities and the layer thicknesses of
            the n parallel layers of the composite wall." [ref]_

        .. [ref] Tsilingiris, P. T. (2004). On the thermal time constant of
            structural walls. Applied Thermal Engineering, 24(5–6), 743–757.
            https://doi.org/10.1016/j.applthermaleng.2003.10.015
        """
        return (1 / self.total_thickness) * sum(
            [
                layer.Material.Density * layer.Material.SpecificHeat * layer.Thickness
                for layer in self.Layers
            ]
        )

    @property
    def specific_heat(self):
        """Get the construction specific heat weighted by wall area mass [J/(kg⋅K)]."""
        return np.average(
            [layer.specific_heat for layer in self.Layers],
            weights=[layer.Thickness * layer.Material.Density for layer in self.Layers],
        )

    @property
    def heat_capacity_per_unit_wall_area(self):
        """Get the construction heat capacity per unit wall area [J/(m2⋅K)].

        Hint:
            :math:`(HC/A)=ρ·c·δ`, where :math:`δ` is the wall thickness.
        """
        return sum([layer.heat_capacity for layer in self.Layers])

    @property
    def total_thickness(self):
        """Get the construction total thickness [m]."""
        return sum([layer.Thickness for layer in self.Layers])

    @property
    def mass_per_unit_area(self):
        """Get the construction mass per unit area [kg/m2]."""
        return sum([layer.Thickness * layer.Material.Density for layer in self.Layers])

    @property
    def time_constant_per_unit_area(self):
        """Get the construction time constant per unit area [seconds/m2]."""
        return self.mass_per_unit_area * self.specific_heat / self.u_factor

    @property
    def solar_reflectance_index(self):
        """Get the Solar Reflectance Index of the exposed surface.

        Hint:
            calculation from K-12 AEDG, derived from ASTM E1980 assuming medium wind
            speed.

        """
        exposed_material = self.Layers[0]  # 0-th layer is exterior layer
        solar_absorptance = exposed_material.Material.SolarAbsorptance
        thermal_emissivity = exposed_material.Material.ThermalEmittance

        x = (20.797 * solar_absorptance - 0.603 * thermal_emissivity) / (
            9.5205 * thermal_emissivity + 12.0
        )
        sri = 123.97 - 141.35 * x + 9.6555 * x * x

        return sri

    def infer_insulation_layer(self):
        """Return the material layer index that corresponds to the insulation layer."""
        return self.Layers.index(max(self.Layers, key=lambda x: x.r_value))

    def combine(self, other, method="dominant_wall", allow_duplicates=False):
        """Combine two OpaqueConstruction together.

        Args:
            other (OpaqueConstruction): The other OpaqueConstruction object to
                combine with.
            method (str): Equivalent wall assembly method. Only 'dominant_wall'
                is safe to use. 'constant_ufactor' is still weird in terms of
                respecting the thermal response of the walls and may cause
                conversion issues with Conduction Transfer Functions (CTFs) in
                EnergyPlus.

        Returns:
            (OpaqueConstruction): the combined ZoneLoad object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        weights = [self.area, other.area]

        meta = self._get_predecessors_meta(other)
        # thicknesses & materials for self
        if method == "constant_ufactor":
            new_m, new_t = self._constant_ufactor(other, weights)
        elif method == "dominant_wall":
            # simply return the dominant wall construction
            oc = self.dominant_wall(other, weights)
            return oc
        else:
            raise ValueError(
                'Possible choices are ["constant_ufactor", "dominant_wall"]'
            )
        # layers for the new OpaqueConstruction
        layers = [MaterialLayer(mat, t) for mat, t in zip(new_m, new_t)]
        new_obj = self.__class__(**meta, Layers=layers)
        new_name = (
            "Combined Opaque Construction {{{}}} with u_value "
            "of {:,.3f} W/m2k".format(uuid.uuid1(), new_obj.u_value)
        )
        new_obj.rename(new_name)
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        new_obj.area = sum(weights)
        return new_obj

    def dominant_wall(self, other, weights):
        """Return dominant wall construction between self and other.

        Args:
            other:
            weights:
        """
        oc = [
            x
            for _, x in sorted(
                zip([2, 1], [self, other]), key=lambda pair: pair[0], reverse=True
            )
        ][0]
        return oc

    def _constant_ufactor(self, other, weights=None):
        """Return materials and thicknesses for constant u-value.

        The constant u-factor method will produce an assembly that has the
        same u-value as an equivalent wall (weighted by wall area) but with a
        mixture of all unique layer materials

        Args:
            other (OpaqueConstruction): The other Construction.
            weights (array_like, optional): An array of weights associated with
                the self and other. Each value contributes to the average
                according to its associated weight. If `weights=None` , then all
                data are assumed to have a weight equal to one.
        """
        from scipy.optimize import minimize

        def obj_func(
            thicknesses,
            materials,
            expected_u_value,
            expected_specific_heat,
            expected_total_thickness,
        ):
            """Objective function for thickness evaluation."""
            u_value = 1 / sum(
                [
                    thickness / mat.Conductivity
                    for thickness, mat in zip(thicknesses, materials)
                ]
            )

            # Specific_heat: (J/kg K)
            h_calc = [
                mat.SpecificHeat for thickness, mat in zip(thicknesses, materials)
            ]

            # (kg/m3) x (m) = (kg/m2)
            mass_per_unit_area = [
                mat.Density * thickness
                for thickness, mat in zip(thicknesses, materials)
            ]
            specific_heat = np.average(h_calc, weights=mass_per_unit_area)
            return (
                (u_value - expected_u_value) ** 2
                + (specific_heat - expected_specific_heat) ** 2
                + (sum(thicknesses) - expected_total_thickness) ** 2
            )

        # U_eq is the weighted average of the wall u_values by their respected total
        # thicknesses. Here, the U_value does not take into account the convective heat
        # transfer coefficients.
        u_equivalent = np.average(
            [self.u_value, other.u_value],
            weights=[self.total_thickness, other.total_thickness],
        )

        # Get a set of all materials sorted by Material Density (descending order)
        materials = list(
            sorted(
                set(
                    [layer.Material for layer in self.Layers]
                    + [layer.Material for layer in other.Layers]
                ),
                key=lambda x: x.Density,
                reverse=True,
            )
        )

        # Setup weights
        if not weights:
            weights = [1.0, 1.0]

        # If weights is a list of zeros. This weight is used in the
        if not np.array(weights).any():
            weights = [1, 1]

        # Calculate the desired equivalent specific heat
        equi_spec_heat = np.average(
            [self.specific_heat, other.specific_heat], weights=weights
        )
        two_wall_thickness = np.average(
            [self.total_thickness, other.total_thickness], weights=weights
        )
        x0 = np.ones(len(materials))
        bnds = tuple([(0.003, None) for layer in materials])
        res = minimize(
            obj_func,
            x0,
            args=(materials, u_equivalent, equi_spec_heat, two_wall_thickness),
            bounds=bnds,
        )

        return np.array(materials), res.x

    @classmethod
    def from_dict(cls, data, materials, **kwargs):
        """Create an OpaqueConstruction from a dictionary.

        Args:
            data (dict): The python dictionary.
            materials (dict): A dictionary of materials with their id as keys.

        .. code-block:: python

            materials = {}  # dict of materials.
            data = {
                 "$id": "140300770659680",
                 "Layers": [
                  {
                   "Material": {
                    "$ref": "140300653743792"
                   },
                   "Thickness": 0.013
                  },
                  {
                   "Material": {
                    "$ref": "140300653743792"
                   },
                   "Thickness": 0.013
                  }
                 ],
                 "AssemblyCarbon": 0.0,
                 "AssemblyCost": 0.0,
                 "AssemblyEnergy": 0.0,
                 "DisassemblyCarbon": 0.0,
                 "DisassemblyEnergy": 0.0,
                 "Category": "Partition",
                 "Comments": "",
                 "DataSource": "ASHRAE 90.1-2007",
                 "Name": "90.1-2007 Nonres 6A Int Wall"
            }

        """
        # resolve Material objects from ref
        layers = [
            MaterialLayer(
                Material=materials[layer["Material"]["$ref"]],
                Thickness=layer["Thickness"],
            )
            for layer in data.pop("Layers")
        ]
        _id = data.pop("$id")
        oc = cls(Layers=layers, id=_id, **data, **kwargs)

        return oc

    @classmethod
    def generic_internalmass(cls, **kwargs):
        """Create a generic internal mass object.

        Args:
            **kwargs: keywords passed to the class constructor.
        """
        mat = OpaqueMaterial(
            Name="Wood 6inch",
            Roughness="MediumSmooth",
            Thickness=0.15,
            Conductivity=0.12,
            Density=540,
            SpecificHeat=1210,
            ThermalAbsorptance=0.7,
            VisibleAbsorptance=0.7,
        )
        return OpaqueConstruction(
            Name="InternalMass",
            Layers=[MaterialLayer(Material=mat, Thickness=0.15)],
            Category="InternalMass",
            **kwargs,
        )

    @classmethod
    def from_epbunch(cls, epbunch, **kwargs):
        """Create an OpaqueConstruction object from an epbunch.

        Possible keys are "BuildingSurface:Detailed" or "InternalMass"

        Args:
            epbunch (EpBunch): The epbunch object.
            **kwargs: keywords passed to the LayeredConstruction constructor.
        """
        assert epbunch.key.lower() in ("internalmass", "construction"), (
            f"Expected ('Internalmass', 'Construction')." f"Got '{epbunch.key}'."
        )
        name = epbunch.Name

        # treat internalmass and regular surfaces differently
        if epbunch.key.lower() == "internalmass":
            layers = cls._internalmass_layer(epbunch)
            return cls(Name=name, Layers=layers, **kwargs)
        elif epbunch.key.lower() == "construction":
            layers = cls._surface_layers(epbunch)
            return cls(Name=name, Layers=layers, **kwargs)

    @classmethod
    def _internalmass_layer(cls, epbunch):
        """Return layers of an internal mass object.

        Args:
            epbunch (EpBunch): The InternalMass epobject.
        """
        constr_obj = epbunch.theidf.getobject("CONSTRUCTION", epbunch.Construction_Name)
        return cls._surface_layers(constr_obj)

    @classmethod
    def _surface_layers(cls, epbunch):
        """Retrieve layers for the OpaqueConstruction.

        Args:
            epbunch (EpBunch): EP-Construction object
        """
        layers = []
        for layer in epbunch.fieldnames[2:]:
            # Iterate over the construction's layers
            material = epbunch.get_referenced_object(layer)
            if material:
                o = OpaqueMaterial.from_epbunch(material, allow_duplicates=True)
                try:
                    thickness = material.Thickness
                except BadEPFieldError:
                    thickness = o.Conductivity * material.Thermal_Resistance
                layers.append(MaterialLayer(Material=o, Thickness=thickness))
        return layers

    def to_dict(self):
        """Return OpaqueConstruction dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Layers"] = [lay.to_dict() for lay in self.Layers]
        data_dict["AssemblyCarbon"] = self.AssemblyCarbon
        data_dict["AssemblyCost"] = self.AssemblyCost
        data_dict["AssemblyEnergy"] = self.AssemblyEnergy
        data_dict["DisassemblyCarbon"] = self.DisassemblyCarbon
        data_dict["DisassemblyEnergy"] = self.DisassemblyEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = str(self.DataSource)
        data_dict["Name"] = self.Name

        return data_dict

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Layers=self.Layers,
            AssemblyCarbon=self.AssemblyCarbon,
            AssemblyCost=self.AssemblyCost,
            AssemblyEnergy=self.AssemblyEnergy,
            DisassemblyCarbon=self.DisassemblyCarbon,
            DisassemblyEnergy=self.DisassemblyEnergy,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    @classmethod
    def generic(cls, **kwargs):
        """Return OpaqueConstruction based on 90.1-2007 Nonres 4B Int Wall."""
        om = OpaqueMaterial.generic()

        layers = [MaterialLayer(om, 0.0127), MaterialLayer(om, 0.0127)]  # half inch
        return cls(
            Name="90.1-2007 Nonres 6A Int Wall",
            Layers=layers,
            DataSource="ASHRAE 90.1-2007",
            Category="Partition",
            **kwargs,
        )

    def __add__(self, other):
        """Overload + to implement self.combine.

        Args:
            other (OpaqueConstruction): The other OpaqueConstruction.
        """
        return self.combine(other)

    def __hash__(self):
        """Return the hash value of self."""
        return hash((self.__class__.__name__, getattr(self, "Name", None)))

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, OpaqueConstruction):
            return NotImplemented
        else:
            return all([self.Layers == other.Layers])

    def __copy__(self):
        """Create a copy of self."""
        new_con = self.__class__(Name=self.Name, Layers=[a for a in self.Layers])
        return new_con

    def to_epbunch(self, idf):
        """Get a Construction EpBunch given an idf model.

        Notes:
            Will create layered materials as well.

        Args:
            idf (IDF): An idf model to add the EpBunch in.

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return idf.newidfobject(
            key="CONSTRUCTION",
            Name=self.Name,
            Outside_Layer=self.Layers[0].to_epbunch(idf).Name,
            **{
                f"Layer_{i+2}": layer.to_epbunch(idf).Name
                for i, layer in enumerate(self.Layers[1:])
            },
        )



################################################################################
# Module: plot.py
# Description: Plot energy profiles, spatial geometries, networks, and routes
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################
# OSMnx
#
# Copyright (c) 2019 Geoff Boeing https://geoffboeing.com/
#
# Part of the following code is a derivative work of the code from the OSMnx
# project, which is licensed MIT License. This code therefore is also
# licensed under the terms of the The MIT License (MIT).
################################################################################

import os
import time

import numpy as np

from archetypal import settings
from archetypal.utils import log


def save_and_show(
    fig, ax, save, show, close, filename, file_format, dpi, axis_off, extent
):
    """Save a figure to disk and show it, as specified.

    Args:
        extent (str or `~matplotlib.transforms.Bbox`, optional): Bbox in
            inches. Only the given portion of the figure is saved. If
            'tight', try to figure out the tight bbox of the figure. If None,
            use savefig.bbox
        fig (matplotlib.figure.Figure): the figure
        ax (matplotlib.axes.Axes): the axes
        save (bool): whether to save the figure to disk or not
        show (bool): whether to display the figure or not
        close (bool): close the figure (only if show equals False) to prevent
            display
        filename (string): the name of the file to save
        file_format (str): the format of the file to save (e.g., 'jpg',
            'png', 'svg')
        dpi (int): the resolution of the image file if saving (Dots per inch)
        axis_off (bool): if True matplotlib axis was turned off by plot_graph so
            constrain the saved figure's extent to the interior of the axis

    Returns:
        (tuple) fig, ax
    """
    import matplotlib.pyplot as plt

    if save:
        start_time = time.time()

        # create the save folder if it doesn't already exist
        if not os.path.exists(settings.imgs_folder):
            os.makedirs(settings.imgs_folder)
        path_filename = os.path.join(
            settings.imgs_folder, os.extsep.join([filename, file_format])
        )

        if not isinstance(ax, (np.ndarray, list)):
            ax = [ax]
        if file_format == "svg":
            fig.patch.set_alpha(0.0)
            fig.savefig(
                path_filename,
                bbox_inches=extent,
                format=file_format,
                facecolor=fig.get_facecolor(),
                transparent=True,
            )
        else:
            if extent is None:
                if len(ax) == 1:
                    if axis_off:
                        for ax in ax:
                            # if axis is turned off, constrain the saved
                            # figure's extent to the interior of the axis
                            extent = ax.get_window_extent().transformed(
                                fig.dpi_scale_trans.inverted()
                            )
                else:
                    pass
            fig.savefig(
                path_filename,
                dpi=dpi,
                bbox_inches=extent,
                format=file_format,
                facecolor=fig.get_facecolor(),
                transparent=True,
            )
        log(
            "Saved the figure to disk in {:,.2f} seconds".format(
                time.time() - start_time
            )
        )

    # show the figure if specified
    if show:
        start_time = time.time()
        plt.show()
        # fig.show()
        log("Showed the plot in {:,.2f} seconds".format(time.time() - start_time))
    # if show=False, close the figure if close=True to prevent display
    elif close:
        plt.close()

    return fig, ax



"""archetypal materials."""

__all__ = (
    "GasMaterial",
    "GlazingMaterial",
    "OpaqueMaterial",
)

from archetypal.template.materials.gas_material import GasMaterial
from archetypal.template.materials.glazing_material import GlazingMaterial
from archetypal.template.materials.opaque_material import OpaqueMaterial



"""archetypal ConstructionBase and LayeredConstruction.

Notes:
    Thank you to `honeybee-energy` for implementing heat transfer coefficient
    formulas from ISO. Those where adapted to the structure of the
    archetypal.template module.
"""

import math
from typing import List, Union

from validator_collection import validators

from archetypal.template.materials import GasMaterial
from archetypal.template.materials.gas_layer import GasLayer
from archetypal.template.materials.material_layer import MaterialLayer
from archetypal.template.umi_base import UmiBase


class ConstructionBase(UmiBase):
    """A class used to store data linked to Life Cycle aspects.

    For more information on the Life Cycle Analysis performed in UMI, see:
    https://umidocs.readthedocs.io/en/latest/docs/life-cycle-introduction.html#life-cycle-impact
    """

    __slots__ = (
        "_assembly_carbon",
        "_assembly_cost",
        "_assembly_energy",
        "_dissassembly_carbon",
        "_dissassembly_energy",
    )

    def __init__(
        self,
        Name,
        AssemblyCarbon=0,
        AssemblyCost=0,
        AssemblyEnergy=0,
        DisassemblyCarbon=0,
        DisassemblyEnergy=0,
        **kwargs,
    ):
        """Initialize a ConstructionBase object with parameters.

        Args:
            AssemblyCarbon (float): assembly carbon [kgCO2/m2].
            AssemblyCost (float): assembly carbon [kgCO2/m2].
            AssemblyEnergy (float): assembly energy [MJ/m2].
            DisassemblyCarbon (float): disassembly carbon [kgCO2/m2].
            DisassemblyEnergy (float): disassembly energy [MJ/m2].
            **kwargs: keywords passed to UmiBase.
        """
        super(ConstructionBase, self).__init__(Name, **kwargs)
        self.AssemblyCarbon = AssemblyCarbon
        self.AssemblyCost = AssemblyCost
        self.AssemblyEnergy = AssemblyEnergy
        self.DisassemblyCarbon = DisassemblyCarbon
        self.DisassemblyEnergy = DisassemblyEnergy

    @property
    def AssemblyCarbon(self):
        """Get or set the assembly carbon [kgCO2/m2]."""
        return self._assembly_carbon

    @AssemblyCarbon.setter
    def AssemblyCarbon(self, value):
        self._assembly_carbon = float(value)

    @property
    def AssemblyCost(self):
        """Get or set the assembly cost [$/m2]."""
        return self._assembly_cost

    @AssemblyCost.setter
    def AssemblyCost(self, value):
        self._assembly_cost = float(value)

    @property
    def AssemblyEnergy(self):
        """Get or set the assembly energy [MJ/m2]."""
        return self._assembly_energy

    @AssemblyEnergy.setter
    def AssemblyEnergy(self, value):
        self._assembly_energy = float(value)

    @property
    def DisassemblyCarbon(self):
        """Get or set the disassembly carbon [kgCO2/m2]."""
        return self._dissassembly_carbon

    @DisassemblyCarbon.setter
    def DisassemblyCarbon(self, value):
        self._dissassembly_carbon = float(value)

    @property
    def DisassemblyEnergy(self):
        """Get or set the disassembly energy [MJ/m2]."""
        return self._dissassembly_energy

    @DisassemblyEnergy.setter
    def DisassemblyEnergy(self, value):
        self._dissassembly_energy = float(value)

    def validate(self):
        """Validate object and fill in missing values."""
        return self

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __key__(self):
        """Get a tuple of attributes. Useful for hashing and comparing."""
        return (
            self.AssemblyCarbon,
            self.AssemblyCost,
            self.AssemblyEnergy,
            self.DisassemblyCarbon,
            self.DisassemblyEnergy,
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        return isinstance(other, ConstructionBase) and self.__key__() == other.__key__()

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(
            self.Name,
            self.AssemblyCarbon,
            self.AssemblyCost,
            self.AssemblyEnergy,
            self.DisassemblyCarbon,
            self.DisassemblyEnergy,
        )


class LayeredConstruction(ConstructionBase):
    """Defines the layers of an :class:`OpaqueConstruction`.

    Attributes:
        Layers (list of archetypal.MaterialLayer): List of MaterialLayer objects from
            outside to inside.
    """

    __slots__ = ("_layers",)

    def __init__(self, Name, Layers, **kwargs):
        """Initialize Layered Construction.

        Args:
            Layers (list of (MaterialLayer or GasLayer)): A list of
                :class:`MaterialLayer` or :class:`GasLayer` objects.
            **kwargs: Keywords passed to the :class:`ConstructionBase`
                constructor.
        """
        super(LayeredConstruction, self).__init__(Name, **kwargs)
        self.Layers = Layers

    @property
    def Layers(self) -> List[Union[MaterialLayer, GasLayer]]:
        """Get or set the material layers."""
        return self._layers

    @Layers.setter
    def Layers(self, value):
        value = validators.iterable(value, minimum_length=1, maximum_length=10)
        assert all(isinstance(a, (MaterialLayer, GasLayer)) for a in value), (
            "Value input error for '{value}'. Layers must be a list of MaterialLayer "
            "or GasLayer objects only."
        )
        assert isinstance(
            value[0], MaterialLayer
        ), "The outside layer cannot be a GasLayer"
        assert isinstance(
            value[-1], MaterialLayer
        ), "The inside layer cannot be a GasLayer"
        self._layers = value

    @property
    def r_value(self):
        """Get or set the thermal resistance [K⋅m2/W] (excluding air films)."""
        return sum([layer.r_value for layer in self.Layers])

    @property
    def u_value(self):
        """Get the heat transfer coefficient [W/m2⋅K] (excluding air films)."""
        return 1 / self.r_value

    @property
    def r_factor(self):
        """Get the R-factor [m2-K/W] (including air films)."""
        return 1 / self.out_h_simple() + self.r_value + 1 / self.in_h_simple()

    def out_h_simple(self):
        """Get the simple outdoor heat transfer coefficient according to ISO 10292.

        This is used for all opaque R-factor calculations.
        """
        return 23

    def in_h_simple(self):
        """Get the simple indoor heat transfer coefficient according to ISO 10292.

        This is used for all opaque R-factor calculations.
        """
        return 3.6 + (4.4 * self.inside_emissivity / 0.84)

    def out_h(self, wind_speed=6.7, t_kelvin=273.15):
        """Get the detailed outdoor heat transfer coefficient according to ISO 15099.

        This is used for window U-factor calculations and all of the
        temperature_profile calculations.

        Args:
            wind_speed (float): The average outdoor wind speed [m/s]. This affects the
                convective heat transfer coefficient. Default is 6.7 m/s.
            t_kelvin (float): The average between the outdoor temperature and the
                exterior surface temperature. This can affect the radiative heat
                transfer. Default is 273.15K (0C).
        """
        _conv_h = 4 + (4 * wind_speed)
        _rad_h = 4 * 5.6697e-8 * self.outside_emissivity * (t_kelvin ** 3)
        return _conv_h + _rad_h

    def in_h(self, t_kelvin=293.15, delta_t=15, height=1.0, angle=90, pressure=101325):
        """Get the detailed indoor heat transfer coefficient according to ISO 15099.

        This is used for window U-factor calculations and all of the
        temperature_profile calculations.

        Args:
            t_kelvin (float): The average between the indoor temperature and the
                interior surface temperature. Default is 293.15K (20C).
            delta_t (float): The temperature difference between the indoor temperature
                and the interior surface temperature [C]. Default is 15C.
            height (float): An optional height for the surface in meters. Default is
                1.0 m, which is consistent with NFRC standards.
            angle (float): An angle in degrees between 0 and 180.
                0 = A horizontal surface with downward heat flow through the layer.
                90 = A vertical surface
                180 = A horizontal surface with upward heat flow through the layer.
            pressure (float): The average pressure in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        _conv_h = self.in_h_c(t_kelvin, delta_t, height, angle, pressure)
        _rad_h = 4 * 5.6697e-8 * self.inside_emissivity * (t_kelvin ** 3)
        return _conv_h + _rad_h

    def in_h_c(
        self, t_kelvin=293.15, delta_t=15, height=1.0, angle=90, pressure=101325
    ):
        """Get detailed indoor convective heat transfer coef. according to ISO 15099.

        This is used for window U-factor calculations and all of the
        temperature_profile calculations.

        Args:
            t_kelvin (float): The average between the indoor temperature and the
                interior surface temperature. Default is 293.15K (20C).
            delta_t (float): The temperature difference between the indoor temperature
                and the interior surface temperature [C]. Default is 15C.
            height (float): An optional height for the surface in meters. Default is
                1.0 m, which is consistent with NFRC standards.
            angle (float): An angle in degrees between 0 and 180.
                0 = A horizontal surface with downward heat flow through the layer.
                90 = A vertical surface
                180 = A horizontal surface with upward heat flow through the layer.
            pressure (float): The average pressure in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        gas_material = GasMaterial("AIR")
        _ray_numerator = (
            (gas_material.density_at_temperature(t_kelvin, pressure) ** 2)
            * (height ** 3)
            * 9.81
            * gas_material.specific_heat_at_temperature(t_kelvin, pressure)
            * delta_t
        )
        _ray_denominator = (
            t_kelvin
            * gas_material.viscosity_at_temperature(t_kelvin, pressure)
            * gas_material.conductivity_at_temperature(t_kelvin, pressure)
        )
        _rayleigh_h = abs(_ray_numerator / _ray_denominator)
        if angle < 15:
            nusselt = 0.13 * (_rayleigh_h ** (1 / 3))
        elif angle <= 90:
            _sin_a = math.sin(math.radians(angle))
            _rayleigh_c = 2.5e5 * ((math.exp(0.72 * angle) / _sin_a) ** (1 / 5))
            if _rayleigh_h < _rayleigh_c:
                nusselt = 0.56 * ((_rayleigh_h * _sin_a) ** (1 / 4))
            else:
                nu_1 = 0.56 * ((_rayleigh_c * _sin_a) ** (1 / 4))
                nu_2 = 0.13 * ((_rayleigh_h ** (1 / 3)) - (_rayleigh_c ** (1 / 3)))
                nusselt = nu_1 + nu_2
        elif angle <= 179:
            _sin_a = math.sin(math.radians(angle))
            nusselt = 0.56 * ((_rayleigh_h * _sin_a) ** (1 / 4))
        else:
            nusselt = 0.58 * (_rayleigh_h ** (1 / 5))
        _conv_h = nusselt * (
            gas_material.conductivity_at_temperature(t_kelvin, pressure) / height
        )
        return _conv_h

    @property
    def outside_emissivity(self):
        """Get the hemispherical emissivity of the outside face of the construction."""
        return self.Layers[0].Material.ThermalEmittance

    @property
    def inside_emissivity(self):
        """Get the emissivity of the inside face of the construction [-]."""
        return self.Layers[-1].Material.ThermalEmittance

    @property
    def u_factor(self):
        """Get the overall heat transfer coefficient (including air films) W/(m2⋅K)."""
        return 1 / self.r_factor

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(Name=self.Name, Layers=self.Layers)

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        return isinstance(other, LayeredConstruction) and all(
            [self.Layers == other.Layers]
        )



"""archetypal GlazingMaterial."""

import collections

from sigfig import round
from validator_collection import validators

from archetypal.idfclass.extensions import EpBunch
from archetypal.template.materials.material_base import MaterialBase
from archetypal.template.umi_base import UmiBase
from archetypal.utils import log


class GlazingMaterial(MaterialBase):
    """Glazing Materials class.

    .. image:: ../images/template/materials-glazing.png

    """

    __slots__ = (
        "_ir_emissivity_back",
        "_ir_emissivity_front",
        "_ir_transmittance",
        "_visible_reflectance_back",
        "_visible_reflectance_front",
        "_visible_transmittance",
        "_solar_reflectance_back",
        "_solar_reflectance_front",
        "_solar_transmittance",
        "_dirt_factor",
        "_conductivity",
    )

    def __init__(
        self,
        Name,
        Density=2500,
        Conductivity=0.9,
        SolarTransmittance=0,
        SolarReflectanceFront=0,
        SolarReflectanceBack=0,
        VisibleTransmittance=0,
        VisibleReflectanceFront=0,
        VisibleReflectanceBack=0,
        IRTransmittance=0,
        IREmissivityFront=0,
        IREmissivityBack=0,
        DirtFactor=1.0,
        Cost=0.0,
        **kwargs,
    ):
        """Initialize a GlazingMaterial object.

        Args:
            Name (str): The name of the GlazingMaterial.
            Density (float): A number representing the density of the material
                in kg/m3. This is essentially the mass of one cubic meter of the
                material.
            Conductivity (float): Thermal conductivity (W/m-K).
            SolarTransmittance (float): Transmittance at normal incidence
                averaged over the solar spectrum.
            SolarReflectanceFront (float): Front-side reflectance at normal
                incidence averaged over the solar spectrum.
            SolarReflectanceBack (float): Back-side reflectance at normal
                incidence averaged over the solar spectrum.
            VisibleTransmittance (float): Transmittance at normal incidence
                averaged over the solar spectrum and weighted by the response of
                the human eye.
            VisibleReflectanceFront (float): Front-side reflectance at normal
                incidence averaged over the solar spectrum and weighted by the
                response of the human eye.
            VisibleReflectanceBack (float): Back-side reflectance at normal
                incidence averaged over the solar spectrum and weighted by the
                response of the human eye.
            IRTransmittance (float): Long-wave transmittance at normal
                incidence.
            IREmissivityFront (float): Front-side long-wave emissivity.
            IREmissivityBack (float): Back-side long-wave emissivity.
            DirtFactor (float): This is a factor that corrects for the presence
                of dirt on the glass. Using a material with dirt correction
                factor < 1.0 in the construction for an interior window will
                result in an error message.
            **kwargs: keywords passed to the :class:`MaterialBase`
                constructor. For more info, see :class:`MaterialBase`.
        """
        super(GlazingMaterial, self).__init__(Name, Cost=Cost, **kwargs)

        self._solar_reflectance_front = 0
        self._solar_reflectance_back = None
        self._visible_reflectance_front = 0
        self._visible_reflectance_back = None

        self.Conductivity = Conductivity
        self.Density = Density
        self.DirtFactor = DirtFactor
        self.IREmissivityBack = IREmissivityBack
        self.IREmissivityFront = IREmissivityFront
        self.IRTransmittance = IRTransmittance
        self.VisibleReflectanceBack = VisibleReflectanceBack
        self.VisibleReflectanceFront = VisibleReflectanceFront
        self.VisibleTransmittance = VisibleTransmittance
        self.SolarReflectanceBack = SolarReflectanceBack
        self.SolarReflectanceFront = SolarReflectanceFront
        self.SolarTransmittance = SolarTransmittance

    @property
    def Conductivity(self):
        """Get or set the conductivity of the material [W/m-K]."""
        return self._conductivity

    @Conductivity.setter
    def Conductivity(self, value):
        self._conductivity = validators.float(value, minimum=0)

    @property
    def Density(self):
        """Get or set the density of the material [J/kg-K]."""
        return self._density

    @Density.setter
    def Density(self, value):
        self._density = validators.float(value, minimum=0)

    @property
    def DirtFactor(self):
        """Get or set the dirt correction factor [-]."""
        return self._dirt_factor

    @DirtFactor.setter
    def DirtFactor(self, value):
        if value == "":
            value = 1
        self._dirt_factor = validators.float(value, minimum=0, maximum=1)

    @property
    def IREmissivityBack(self):
        """Get or set the infrared emissivity of the back side [-]."""
        return float(self._ir_emissivity_back)

    @IREmissivityBack.setter
    def IREmissivityBack(self, value):
        self._ir_emissivity_back = validators.float(value, False, 0.0, 1.0)

    @property
    def IREmissivityFront(self):
        """Get or set the infrared emissivity of the front side [-]."""
        return self._ir_emissivity_front

    @IREmissivityFront.setter
    def IREmissivityFront(self, value):
        self._ir_emissivity_front = validators.float(value, False, 0.0, 1.0)

    @property
    def IRTransmittance(self):
        """Get or set the infrared transmittance [-]."""
        return self._ir_transmittance

    @IRTransmittance.setter
    def IRTransmittance(self, value):
        self._ir_transmittance = validators.float(value, False, 0.0, 1.0)

    @property
    def VisibleReflectanceBack(self):
        """Get or set the visible reflectance of the back side [-]."""
        return self._visible_reflectance_back

    @VisibleReflectanceBack.setter
    def VisibleReflectanceBack(self, value):
        self._visible_reflectance_back = validators.float(value, False, 0.0, 1.0)

    @property
    def VisibleReflectanceFront(self):
        """Get or set the visible reflectance of the front side [-]."""
        return self._visible_reflectance_front

    @VisibleReflectanceFront.setter
    def VisibleReflectanceFront(self, value):
        self._visible_reflectance_front = validators.float(value, False, 0.0, 1.0)

    @property
    def VisibleTransmittance(self):
        """Get or set the visible transmittance [-]."""
        return self._visible_transmittance

    @VisibleTransmittance.setter
    def VisibleTransmittance(self, value):
        assert value + self._visible_reflectance_front <= 1, (
            f"Sum of window transmittance and reflectance '"
            f"{self._visible_reflectance_front}' is greater than 1."
        )
        if self._visible_reflectance_back is not None:
            assert value + self._visible_reflectance_back <= 1, (
                f"Sum of window transmittance and reflectance '"
                f"{self._visible_reflectance_back}' is greater than 1."
            )
        self._visible_transmittance = validators.float(value, False, 0.0, 1.0)

    @property
    def SolarReflectanceBack(self):
        """Get or set the solar reflectance of the back side [-]."""
        return self._solar_reflectance_back

    @SolarReflectanceBack.setter
    def SolarReflectanceBack(self, value):
        self._solar_reflectance_back = validators.float(value, False, 0.0, 1.0)

    @property
    def SolarReflectanceFront(self):
        """Get or set the solar reflectance of the front side [-]."""
        return self._solar_reflectance_front

    @SolarReflectanceFront.setter
    def SolarReflectanceFront(self, value):
        self._solar_reflectance_front = validators.float(value, False, 0.0, 1.0)

    @property
    def SolarTransmittance(self):
        """Get or set the solar transmittance [-]."""
        return self._solar_transmittance

    @SolarTransmittance.setter
    def SolarTransmittance(self, value):
        self._solar_transmittance = validators.float(value, False, 0.0, 1.0)

    def combine(self, other, weights=None, allow_duplicates=False):
        """Combine two GlazingMaterial objects together.

        Args:
            other (GlazingMaterial): The other GlazingMaterial object to
                combine with.
        Returns:
            (GlazingMaterial): the combined GlazingMaterial object.
        """
        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        meta = self._get_predecessors_meta(other)

        if not weights:
            log(
                'using GlazingMaterial density as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [self.Density, other.Density]
        # iterate over attributes and apply either float_mean or str_mean.
        new_attr = {}
        for attr, value in self.mapping().items():
            if attr not in ["Comments", "DataSource"]:
                if isinstance(value, (int, float)) or isinstance(other, (int, float)):
                    new_attr[attr] = UmiBase.float_mean(
                        self, other, attr=attr, weights=weights
                    )
                elif isinstance(value, str) or isinstance(other, str):
                    new_attr[attr] = UmiBase._str_mean(
                        self, other, attr=attr, append=False
                    )
                elif isinstance(value, list) or isinstance(other, list):
                    new_attr[attr] = getattr(self, attr) + getattr(other, attr)
                elif isinstance(value, collections.UserList) or isinstance(
                    other, collections.UserList
                ):
                    pass
                else:
                    raise NotImplementedError
        [new_attr.pop(key, None) for key in meta.keys()]  # meta handles these
        # keywords.
        # create a new object from combined attributes
        new_obj = self.__class__(**meta, **new_attr)
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def to_dict(self):
        """Return GlazingMaterial dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["DirtFactor"] = self.DirtFactor
        data_dict["IREmissivityBack"] = round(self.IREmissivityBack, 2)
        data_dict["IREmissivityFront"] = round(self.IREmissivityFront, 2)
        data_dict["IRTransmittance"] = round(self.IRTransmittance, 2)
        data_dict["SolarReflectanceBack"] = round(self.SolarReflectanceBack, 2)
        data_dict["SolarReflectanceFront"] = round(self.SolarReflectanceFront, 2)
        data_dict["SolarTransmittance"] = round(self.SolarTransmittance, 2)
        data_dict["VisibleReflectanceBack"] = round(self.VisibleReflectanceBack, 2)
        data_dict["VisibleReflectanceFront"] = round(self.VisibleReflectanceFront, 2)
        data_dict["VisibleTransmittance"] = round(self.VisibleTransmittance, 2)
        data_dict["Conductivity"] = round(self.Conductivity, 2)
        data_dict["Cost"] = self.Cost
        data_dict["Density"] = self.Density
        data_dict["EmbodiedCarbon"] = self.EmbodiedCarbon
        data_dict["EmbodiedEnergy"] = self.EmbodiedEnergy
        data_dict["SubstitutionRatePattern"] = self.SubstitutionRatePattern
        data_dict["SubstitutionTimestep"] = self.SubstitutionTimestep
        data_dict["TransportCarbon"] = self.TransportCarbon
        data_dict["TransportDistance"] = self.TransportDistance
        data_dict["TransportEnergy"] = self.TransportEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def to_epbunch(self, idf, thickness) -> EpBunch:
        """Convert self to an EpBunch given an idf model and a thickness.

        Args:
            idf (IDF): An IDF model.
            thickness (float): the thickness of the material.

        .. code-block:: python

            WindowMaterial:Glazing,
                B_Glass_Clear_3_0.003_B_Dbl_Air_Cl,    !- Name
                SpectralAverage,          !- Optical Data Type
                SpectralAverage,          !- Window Glass Spectral Data Set Name
                0.003,                    !- Thickness
                0.83,                     !- Solar Transmittance at Normal Incidence
                0.07,                     !- Front Side Solar Reflectance at Normal Incidence
                0.07,                     !- Back Side Solar Reflectance at Normal Incidence
                0.89,                     !- Visible Transmittance at Normal Incidence
                0.08,                     !- Front Side Visible Reflectance at Normal Incidence
                0.08,                     !- Back Side Visible Reflectance at Normal Incidence
                0,                        !- Infrared Transmittance at Normal Incidence
                0.84,                     !- Front Side Infrared Hemispherical Emissivity
                0.84,                     !- Back Side Infrared Hemispherical Emissivity
                0.9,                      !- Conductivity
                1;                        !- Dirt Correction Factor for Solar and Visible Transmittance

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return idf.newidfobject(
            "WINDOWMATERIAL:GLAZING",
            Name=self.Name,
            Optical_Data_Type="SpectralAverage",
            Window_Glass_Spectral_Data_Set_Name="SpectralAverage",
            Thickness=thickness,
            Solar_Transmittance_at_Normal_Incidence=self.SolarTransmittance,
            Front_Side_Solar_Reflectance_at_Normal_Incidence=self.SolarReflectanceFront,
            Back_Side_Solar_Reflectance_at_Normal_Incidence=self.SolarReflectanceBack,
            Visible_Transmittance_at_Normal_Incidence=self.VisibleTransmittance,
            Front_Side_Visible_Reflectance_at_Normal_Incidence=self.VisibleReflectanceFront,
            Back_Side_Visible_Reflectance_at_Normal_Incidence=self.VisibleReflectanceBack,
            Infrared_Transmittance_at_Normal_Incidence=self.IRTransmittance,
            Front_Side_Infrared_Hemispherical_Emissivity=self.IREmissivityFront,
            Back_Side_Infrared_Hemispherical_Emissivity=self.IREmissivityBack,
            Conductivity=self.Conductivity,
            Dirt_Correction_Factor_for_Solar_and_Visible_Transmittance=self.DirtFactor,
        )

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            DirtFactor=self.DirtFactor,
            IREmissivityBack=self.IREmissivityBack,
            IREmissivityFront=self.IREmissivityFront,
            IRTransmittance=self.IRTransmittance,
            SolarReflectanceBack=self.SolarReflectanceBack,
            SolarReflectanceFront=self.SolarReflectanceFront,
            SolarTransmittance=self.SolarTransmittance,
            VisibleReflectanceBack=self.VisibleReflectanceBack,
            VisibleReflectanceFront=self.VisibleReflectanceFront,
            VisibleTransmittance=self.VisibleTransmittance,
            Conductivity=self.Conductivity,
            Cost=self.Cost,
            Density=self.Density,
            EmbodiedCarbon=self.EmbodiedCarbon,
            EmbodiedEnergy=self.EmbodiedEnergy,
            SubstitutionRatePattern=self.SubstitutionRatePattern,
            SubstitutionTimestep=self.SubstitutionTimestep,
            TransportCarbon=self.TransportCarbon,
            TransportDistance=self.TransportDistance,
            TransportEnergy=self.TransportEnergy,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Create a GlazingMaterial from a dictionary.

        Args:
            data: The python dictionary following the format below.

        .. code-block:: python

            {'$id': '1',
             'DirtFactor': 1.0,
             'IREmissivityBack': 0.0,
             'IREmissivityFront': 0.0,
             'IRTransmittance': 0.0,
             'SolarReflectanceBack': 0.0,
             'SolarReflectanceFront': 0.0,
             'SolarTransmittance': 0.0,
             'VisibleReflectanceBack': 0.0,
             'VisibleReflectanceFront': 0.0,
             'VisibleTransmittance': 0.0,
             'Conductivity': 0.0,
             'Cost': 0.0,
             'Density': 2500,
             'EmbodiedCarbon': 0.0,
             'EmbodiedEnergy': 0.0,
             'SubstitutionRatePattern': [1.0],
             'SubstitutionTimestep': 100.0,
             'TransportCarbon': 0.0,
             'TransportDistance': 0.0,
             'TransportEnergy': 0.0,
             'Category': 'Uncategorized',
             'Comments': '',
             'DataSource': None,
             'Name': 'A'}
        """
        _id = data.pop("$id")
        return cls(id=_id, **data, **kwargs)

    def __add__(self, other):
        """Overload + to implement self.combine."""
        return self.combine(other)

    def __hash__(self):
        """Return the hash value of self."""
        return hash((self.__class__.__name__, getattr(self, "Name", None)))

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, GlazingMaterial):
            return NotImplemented
        else:
            return all(
                [
                    self.Density == other.Density,
                    self.Conductivity == other.Conductivity,
                    self.SolarTransmittance == other.SolarTransmittance,
                    self.SolarReflectanceFront == other.SolarReflectanceFront,
                    self.SolarReflectanceBack == other.SolarReflectanceBack,
                    self.VisibleTransmittance == other.VisibleTransmittance,
                    self.VisibleReflectanceFront == other.VisibleReflectanceFront,
                    self.VisibleReflectanceBack == other.VisibleReflectanceBack,
                    self.IRTransmittance == other.IRTransmittance,
                    self.IREmissivityFront == other.IREmissivityFront,
                    self.IREmissivityBack == other.IREmissivityBack,
                    self.DirtFactor == other.DirtFactor,
                    self.Cost == other.Cost,
                ]
            )

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(**self.mapping())



import collections
import json
import os

import pytest
from path import Path

from archetypal import IDF, settings
from archetypal.eplus_interface.version import get_eplus_dirs
from archetypal.template.building_template import BuildingTemplate
from archetypal.template.conditioning import ZoneConditioning
from archetypal.template.dhw import DomesticHotWaterSetting
from archetypal.template.load import ZoneLoad
from archetypal.template.materials.gas_layer import GasLayer
from archetypal.template.materials.gas_material import GasMaterial
from archetypal.template.materials.glazing_material import GlazingMaterial
from archetypal.template.materials.material_layer import MaterialLayer
from archetypal.template.materials.opaque_material import OpaqueMaterial
from archetypal.template.constructions.opaque_construction import OpaqueConstruction
from archetypal.template.schedule import (
    DaySchedule,
    WeekSchedule,
    YearSchedule,
    YearSchedulePart,
)
from archetypal.template.structure import MassRatio, StructureInformation
from archetypal.template.ventilation import VentilationSetting
from archetypal.template.constructions.window_construction import WindowConstruction
from archetypal.template.window_setting import WindowSetting
from archetypal.template.zone_construction_set import ZoneConstructionSet
from archetypal.template.zonedefinition import ZoneDefinition
from archetypal.umi_template import UmiTemplateLibrary, no_duplicates


class TestUmiTemplate:
    """Test suite for the UmiTemplateLibrary class"""

    def test_template_to_template(self):
        """load the json into UmiTemplateLibrary object, then convert back to json and
        compare"""

        file = "tests/input_data/umi_samples/BostonTemplateLibrary_nodup.json"

        a = UmiTemplateLibrary.open(file).to_dict()
        b = TestUmiTemplate.read_json(file)

        for key in b:
            # Sort the list elements by their Name because .to_dict() sorts by Name
            b[key] = sorted(b[key], key=lambda x: x.get("Name"))
        assert json.loads(json.dumps(a)) == json.loads(json.dumps(b))

    def test_umitemplate(self, config):
        """Test creating UmiTemplateLibrary from 2 IDF files"""
        idf_source = [
            "tests/input_data/necb/NECB 2011-FullServiceRestaurant-NECB HDD Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
            get_eplus_dirs(settings.ep_version)
            / "ExampleFiles"
            / "VentilationSimpleTest.idf",
        ]
        wf = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
        a = UmiTemplateLibrary.from_idf_files(
            idf_source, wf, name="Mixed_Files", processors=-1
        )

        data_dict = a.to_dict()
        a.to_dict()
        assert no_duplicates(data_dict)

    @pytest.mark.skipif(
        os.environ.get("CI", "False").lower() == "true",
        reason="not necessary to test this on CI",
    )
    def test_umi_samples(self, config):
        idf_source = [
            "tests/input_data/umi_samples/B_Off_0.idf",
            "tests/input_data/umi_samples/B_Ret_0.idf",
            "tests/input_data/umi_samples/B_Res_0_Masonry.idf",
            "tests/input_data/umi_samples/B_Res_0_WoodFrame.idf",
        ]
        wf = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
        a = UmiTemplateLibrary.from_idf_files(idf_source, wf, name="Mixed_Files")
        a.to_dict()
        data_dict = a.to_dict()
        assert no_duplicates(data_dict)

    @staticmethod
    def read_json(file):
        with open(file, "r") as f:
            a = json.load(f, object_pairs_hook=collections.OrderedDict)
            data_dict = collections.OrderedDict(
                {
                    "GasMaterials": [],
                    "GlazingMaterials": [],
                    "OpaqueMaterials": [],
                    "OpaqueConstructions": [],
                    "WindowConstructions": [],
                    "StructureDefinitions": [],
                    "DaySchedules": [],
                    "WeekSchedules": [],
                    "YearSchedules": [],
                    "DomesticHotWaterSettings": [],
                    "VentilationSettings": [],
                    "ZoneConditionings": [],
                    "ZoneConstructionSets": [],
                    "ZoneLoads": [],
                    "Zones": [],
                    "WindowSettings": [],
                    "BuildingTemplates": [],
                }
            )
            data_dict.update(a)
            for key in data_dict:
                # Sort the list elements by $id
                data_dict[key] = sorted(
                    data_dict[key], key=lambda x: int(x.get("$id", 0))
                )
            return data_dict

    @pytest.fixture()
    def idf(self):
        yield IDF(prep_outputs=False)

    @pytest.fixture()
    def manual_umitemplate_library(self, config):
        """ Creates Umi template from scratch """

        # region Defines materials

        # Opaque materials
        concrete = OpaqueMaterial(
            Name="Concrete",
            Conductivity=0.5,
            SpecificHeat=800,
            Density=1500,
        )
        insulation = OpaqueMaterial(
            Name="Insulation",
            Conductivity=0.04,
            SpecificHeat=1000,
            Density=30,
        )
        brick = OpaqueMaterial(
            Name="Brick",
            Conductivity=1,
            SpecificHeat=900,
            Density=1900,
        )
        plywood = OpaqueMaterial(
            Name="Plywood",
            Conductivity=0.13,
            SpecificHeat=800,
            Density=540,
        )
        OpaqueMaterials = [concrete, insulation, brick, plywood]

        # Glazing materials
        glass = GlazingMaterial(
            Name="Glass",
            Density=2500,
            Conductivity=1,
            SolarTransmittance=0.7,
            SolarReflectanceFront=0.5,
            SolarReflectanceBack=0.5,
            VisibleTransmittance=0.7,
            VisibleReflectanceFront=0.3,
            VisibleReflectanceBack=0.3,
            IRTransmittance=0.7,
            IREmissivityFront=0.5,
            IREmissivityBack=0.5,
        )
        GlazingMaterials = [glass]

        # Gas materials
        air = GasMaterial(Name="Air", Conductivity=0.02, Density=1.24)
        GasMaterials = [air]
        # endregion

        # region Defines MaterialLayers

        # Opaque MaterialLayers
        concreteLayer = MaterialLayer(concrete, Thickness=0.2)
        insulationLayer = MaterialLayer(insulation, Thickness=0.5)
        brickLayer = MaterialLayer(brick, Thickness=0.1)
        plywoodLayer = MaterialLayer(plywood, Thickness=0.016)

        # Glazing MaterialLayers
        glassLayer = MaterialLayer(glass, Thickness=0.16)

        # Gas MaterialLayers
        airLayer = GasLayer(air, Thickness=0.04)

        MaterialLayers = [
            concreteLayer,
            insulationLayer,
            brickLayer,
            plywoodLayer,
            glassLayer,
            airLayer,
        ]
        # endregion

        # region Defines constructions

        # Opaque constructions
        wall_int = OpaqueConstruction(
            Name="wall_int",
            Layers=[plywoodLayer],
        )
        wall_ext = OpaqueConstruction(
            Name="wall_ext",
            Layers=[concreteLayer, insulationLayer, brickLayer],
        )
        floor = OpaqueConstruction(
            Name="floor",
            Layers=[concreteLayer, plywoodLayer],
        )
        roof = OpaqueConstruction(
            Name="roof",
            Layers=[plywoodLayer, insulationLayer, brickLayer],
        )
        OpaqueConstructions = [wall_int, wall_ext, floor, roof]

        # Window construction
        window = WindowConstruction(
            Layers=[glassLayer, airLayer, glassLayer], Name="Window"
        )
        WindowConstructions = [window]

        # Structure definition
        mass_ratio = MassRatio(Material=plywood, NormalRatio=1, HighLoadRatio=1)
        struct_definition = StructureInformation(
            MassRatios=[mass_ratio], Name="Structure"
        )
        StructureDefinitions = [struct_definition]
        # endregion

        # region Defines schedules

        # Day schedules
        # Always on
        sch_d_on = DaySchedule.from_values(
            Name="AlwaysOn", Values=[1] * 24, Type="Fraction", Category="Day"
        )
        # Always off
        sch_d_off = DaySchedule.from_values(
            Name="AlwaysOff", Values=[0] * 24, Type="Fraction", Category="Day"
        )
        # DHW
        sch_d_dhw = DaySchedule.from_values(
            Name="DHW", Values=[0.3] * 24, Type="Fraction", Category="Day"
        )
        # Internal gains
        sch_d_gains = DaySchedule.from_values(
            Name="Gains",
            Values=[0] * 6 + [0.5, 0.6, 0.7, 0.8, 0.9, 1] + [0.7] * 6 + [0.4] * 6,
            Type="Fraction",
            Category="Day",
        )
        DaySchedules = [sch_d_on, sch_d_dhw, sch_d_gains, sch_d_off]

        # Week schedules
        # Always on
        sch_w_on = WeekSchedule(
            Days=[sch_d_on, sch_d_on, sch_d_on, sch_d_on, sch_d_on, sch_d_on, sch_d_on],
            Category="Week",
            Type="Fraction",
            Name="AlwaysOn",
        )
        # Always off
        sch_w_off = WeekSchedule(
            Days=[
                sch_d_off,
                sch_d_off,
                sch_d_off,
                sch_d_off,
                sch_d_off,
                sch_d_off,
                sch_d_off,
            ],
            Category="Week",
            Type="Fraction",
            Name="AlwaysOff",
        )
        # DHW
        sch_w_dhw = WeekSchedule(
            Days=[
                sch_d_dhw,
                sch_d_dhw,
                sch_d_dhw,
                sch_d_dhw,
                sch_d_dhw,
                sch_d_dhw,
                sch_d_dhw,
            ],
            Category="Week",
            Type="Fraction",
            Name="DHW",
        )
        # Internal gains
        sch_w_gains = WeekSchedule(
            Days=[
                sch_d_gains,
                sch_d_gains,
                sch_d_gains,
                sch_d_gains,
                sch_d_gains,
                sch_d_gains,
                sch_d_gains,
            ],
            Category="Week",
            Type="Fraction",
            Name="Gains",
        )
        WeekSchedules = [sch_w_on, sch_w_off, sch_w_dhw, sch_w_gains]

        # Year schedules
        # Always on
        dict_on = {
            "$id": 1,
            "Category": "Year",
            "Parts": [
                {
                    "FromDay": 1,
                    "FromMonth": 1,
                    "ToDay": 31,
                    "ToMonth": 12,
                    "Schedule": sch_w_on.to_ref(),
                }
            ],
            "Type": "Fraction",
            "Name": "AlwaysOn",
        }
        sch_y_on = YearSchedule.from_dict(dict_on, {a.id: a for a in WeekSchedules})
        # Always off
        dict_off = {
            "$id": 2,
            "Category": "Year",
            "Parts": [
                {
                    "FromDay": 1,
                    "FromMonth": 1,
                    "ToDay": 31,
                    "ToMonth": 12,
                    "Schedule": sch_w_off.to_ref(),
                }
            ],
            "Type": "Fraction",
            "Name": "AlwaysOff",
        }
        sch_y_off = YearSchedule.from_dict(dict_off, {a.id: a for a in WeekSchedules})
        # DHW
        dict_dhw = {
            "$id": 3,
            "Category": "Year",
            "Parts": [
                {
                    "FromDay": 1,
                    "FromMonth": 1,
                    "ToDay": 31,
                    "ToMonth": 12,
                    "Schedule": sch_w_dhw.to_ref(),
                }
            ],
            "Type": "Fraction",
            "Name": "DHW",
        }
        sch_y_dhw = YearSchedule.from_dict(dict_dhw, {a.id: a for a in WeekSchedules})
        # Internal gains
        dict_gains = {
            "$id": 4,
            "Category": "Year",
            "Parts": [
                {
                    "FromDay": 1,
                    "FromMonth": 1,
                    "ToDay": 31,
                    "ToMonth": 12,
                    "Schedule": sch_w_gains.to_ref(),
                }
            ],
            "Type": "Fraction",
            "Name": "Gains",
        }
        sch_y_gains = YearSchedule.from_dict(
            dict_gains, {a.id: a for a in WeekSchedules}
        )
        YearSchedules = [sch_y_on, sch_y_off, sch_y_dhw, sch_y_gains]
        # endregion

        # region Defines Window settings

        window_setting = WindowSetting(
            Name="window_setting_1",
            Construction=window,
            AfnWindowAvailability=sch_y_off,
            ShadingSystemAvailabilitySchedule=sch_y_off,
            ZoneMixingAvailabilitySchedule=sch_y_off,
        )
        WindowSettings = [window_setting]
        # endregion

        # region Defines DHW settings

        dhw_setting = DomesticHotWaterSetting(
            WaterSchedule=sch_y_dhw,
            IsOn=True,
            FlowRatePerFloorArea=0.03,
            WaterSupplyTemperature=65,
            WaterTemperatureInlet=10,
            area=1,
            Name="dhw_setting_1",
        )
        DomesticHotWaterSettings = [dhw_setting]
        # endregion

        # region Defines ventilation settings

        vent_setting = VentilationSetting(
            NatVentSchedule=sch_y_off,
            ScheduledVentilationSchedule=sch_y_off,
            Name="vent_setting_1",
        )
        VentilationSettings = [vent_setting]
        # endregion

        # region Defines zone conditioning setttings

        zone_conditioning = ZoneConditioning(
            Name="conditioning_setting_1",
            HeatingSchedule=sch_y_on,
            CoolingSchedule=sch_y_on,
            MechVentSchedule=sch_y_off,
        )
        ZoneConditionings = [zone_conditioning]
        # endregion

        # region Defines zone construction sets

        # Perimeter zone
        zone_constr_set_perim = ZoneConstructionSet(
            Name="constr_set_perim",
            Zone_Names=None,
            Slab=floor,
            IsSlabAdiabatic=False,
            Roof=roof,
            IsRoofAdiabatic=False,
            Partition=wall_int,
            IsPartitionAdiabatic=False,
            Ground=floor,
            IsGroundAdiabatic=False,
            Facade=wall_ext,
            IsFacadeAdiabatic=False,
        )
        # Core zone
        zone_constr_set_core = ZoneConstructionSet(
            Name="constr_set_core",
            Zone_Names=None,
            Slab=floor,
            IsSlabAdiabatic=False,
            Roof=roof,
            IsRoofAdiabatic=False,
            Partition=wall_int,
            IsPartitionAdiabatic=True,
            Ground=floor,
            IsGroundAdiabatic=False,
            Facade=wall_ext,
            IsFacadeAdiabatic=False,
        )
        ZoneConstructionSets = [zone_constr_set_perim, zone_constr_set_core]
        # endregion

        # region Defines zone loads

        zone_load = ZoneLoad(
            EquipmentAvailabilitySchedule=sch_y_gains,
            LightsAvailabilitySchedule=sch_y_gains,
            OccupancySchedule=sch_y_gains,
            Name="zone_load_1",
        )
        ZoneLoads = [zone_load]
        # endregion

        # region Defines zones

        # Perimeter zone
        perim = ZoneDefinition(
            Name="Perim_zone",
            Conditioning=zone_conditioning,
            Constructions=zone_constr_set_perim,
            DomesticHotWater=dhw_setting,
            Loads=zone_load,
            Ventilation=vent_setting,
            Windows=window_setting,
            InternalMassConstruction=wall_int,
        )
        # Core zone
        core = ZoneDefinition(
            Name="Core_zone",
            Conditioning=zone_conditioning,
            Constructions=zone_constr_set_core,
            DomesticHotWater=dhw_setting,
            Loads=zone_load,
            Ventilation=vent_setting,
            Windows=window_setting,
            InternalMassConstruction=wall_int,
        )
        Zones = [perim, core]
        # endregion

        # region Defines building template

        building_template = BuildingTemplate(
            Core=core,
            Perimeter=perim,
            Structure=struct_definition,
            Windows=window_setting,
            Name="Building_template_1",
        )
        BuildingTemplates = [building_template]
        # endregion

        # region Creates json file (Umi template)

        umi_template = UmiTemplateLibrary(
            name="unnamed",
            BuildingTemplates=BuildingTemplates,
            GasMaterials=GasMaterials,
            GlazingMaterials=GlazingMaterials,
            OpaqueConstructions=OpaqueConstructions,
            OpaqueMaterials=OpaqueMaterials,
            WindowConstructions=WindowConstructions,
            StructureInformations=StructureDefinitions,
            DaySchedules=DaySchedules,
            WeekSchedules=WeekSchedules,
            YearSchedules=YearSchedules,
            DomesticHotWaterSettings=DomesticHotWaterSettings,
            VentilationSettings=VentilationSettings,
            WindowSettings=WindowSettings,
            ZoneConditionings=ZoneConditionings,
            ZoneConstructionSets=ZoneConstructionSets,
            ZoneLoads=ZoneLoads,
            ZoneDefinitions=Zones,
        )

        yield umi_template.to_dict()

    def test_manual_template_library(self, manual_umitemplate_library):
        assert no_duplicates(manual_umitemplate_library, attribute="Name")
        assert no_duplicates(manual_umitemplate_library, attribute="$id")

    def test_climatestudio(self, climatestudio):
        template_json = UmiTemplateLibrary(
            name="my_umi_template", BuildingTemplates=[climatestudio]
        ).to_json()
        print(template_json)

    @pytest.mark.skipif(
        os.environ.get("CI", "False").lower() == "true",
        reason="Skipping this test on CI environment",
    )
    @pytest.mark.parametrize(
        "file",
        (
            "tests/input_data/necb/NECB 2011-SmallOffice-NECB HDD Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
            "tests/input_data/necb/NECB 2011-MediumOffice-NECB HDD Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
            "tests/input_data/necb/NECB 2011-LargeOffice-NECB HDD Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
        ),
        ids=("small", "medium", "large"),
    )
    def test_necb_serial(self, file, config):
        settings.log_console = True
        w = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
        template = UmiTemplateLibrary.from_idf_files(
            name="my_umi_template",
            idf_files=[file],
            as_version="9-2-0",
            weather=w,
            processors=1,
        )
        assert no_duplicates(template.to_dict(), attribute="Name")
        assert no_duplicates(template.to_dict(), attribute="$id")

    @pytest.mark.skipif(
        os.environ.get("CI", "False").lower() == "true",
        reason="Skipping this test on CI environment",
    )
    def test_necb_parallel(self, config):
        settings.log_console = True
        office = [
            "tests/input_data/necb/NECB 2011-SmallOffice-NECB HDD "
            "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
            "tests/input_data/necb/NECB 2011-MediumOffice-NECB HDD "
            "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
            "tests/input_data/necb/NECB 2011-LargeOffice-NECB HDD "
            "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
        ]
        w = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
        template = UmiTemplateLibrary.from_idf_files(
            name="my_umi_template",
            idf_files=office,
            as_version="9-2-0",
            weather=w,
            processors=-1,
        )
        template.to_dict()
        assert no_duplicates(template.to_dict(), attribute="Name")
        assert no_duplicates(template.to_dict(), attribute="$id")

    office = [
        "tests/input_data/necb/NECB 2011-SmallOffice-NECB HDD "
        "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
        "tests/input_data/necb/NECB 2011-MediumOffice-NECB HDD "
        "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
        "tests/input_data/necb/NECB 2011-LargeOffice-NECB HDD "
        "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf",
    ]

    @pytest.mark.skipif(
        os.environ.get("CI", "False").lower() == "true",
        reason="Skipping this test on CI environment",
    )
    @pytest.mark.parametrize(
        "file", Path("tests/input_data/problematic").files("*CZ5A*.idf")
    )
    def test_cz5a_serial(self, file, config):
        settings.log_console = True
        w = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
        template = UmiTemplateLibrary.from_idf_files(
            name=file.stem,
            idf_files=[file],
            as_version="9-2-0",
            weather=w,
            processors=1,
        )
        assert no_duplicates(template.to_dict(), attribute="Name")
        assert no_duplicates(template.to_dict(), attribute="$id")


@pytest.fixture(scope="session")
def climatestudio(config):
    """A building template fixture from a climate studio idf file used in subsequent
    tests"""
    file = "tests/input_data/umi_samples/climatestudio_test.idf"
    w = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
    idf = IDF(file, epw=w, annual=True)
    if idf.sim_info is None:
        idf.simulate()

    bt = BuildingTemplate.from_idf(idf)
    yield bt


@pytest.fixture(scope="session")
def sf_cz5a(config):
    """A building template fixture from a climate studio idf file used in subsequent
    tests"""
    file = "tests/input_data/problematic/SF+CZ5A+USA_IL_Chicago-OHare.Intl.AP.725300+oilfurnace+slab+IECC_2012.idf"
    w = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
    idf = IDF(file, epw=w, annual=True)

    bt = BuildingTemplate.from_idf(idf)
    yield bt



class Outputs:
    """Handles preparation of EnergyPlus outputs. Different instance methods
    allow to chain methods together and to add predefined bundles of outputs in
    one go.

    Examples:
        >>> from archetypal import IDF
        >>> idf = IDF(prep_outputs=False)  # True be default
        >>> idf.outputs.add_output_control().add_umi_ouputs(
        >>> ).add_profile_gas_elect_ouputs().apply()
    """

    def __init__(self, idf):
        """Initialize an outputs object.

        Args:
            idf (IDF): the IDF object for wich this outputs object is created.
        """
        self.idf = idf
        self._outputs = []

    def add_custom(self, outputs):
        """Add custom-defined outputs as a list of objects.

        Examples:
            >>> outputs = IDF().outputs
            >>> to_add = dict(
            >>>       key= "OUTPUT:METER",
            >>>       Key_Name="Electricity:Facility",
            >>>       Reporting_Frequency="hourly",
            >>> )
            >>> outputs.add_custom([to_add]).apply()

        Args:
            outputs (list, bool): Pass a list of ep-objects defined as dictionary. See
                examples. If a bool, ignored.

        Returns:
            Outputs: self
        """
        if isinstance(outputs, list):
            self._outputs.extend(outputs)
        return self

    def add_basics(self):
        """Adds the summary report and the sql file to the idf outputs"""
        return (
            self.add_summary_report()
            .add_output_control()
            .add_sql()
            .add_schedules()
            .add_meter_variables()
        )

    def add_schedules(self):
        """Adds Schedules object"""
        outputs = [{"key": "Output:Schedules".upper(), **dict(Key_Field="Hourly")}]

        self._outputs.extend(outputs)
        return self

    def add_meter_variables(self, format="IDF"):
        """Generate .mdd file at end of simulation. This file (from the
        Output:VariableDictionary, regular; and Output:VariableDictionary,
        IDF; commands) shows all the report meters along with their “availability”
        for the current input file. A user must first run the simulation (at least
        semi-successfully) before the available output meters are known. This output
        file is available in two flavors: regular (listed as they are in the Input
        Output Reference) and IDF (ready to be copied and pasted into your Input File).

        Args:
            format (str): Choices are "IDF" and "regul

        Returns:
            Outputs: self
        """
        outputs = [dict(key="Output:VariableDictionary".upper(), Key_Field=format)]
        self._outputs.extend(outputs)
        return self

    def add_summary_report(self, summary="AllSummary"):
        """Adds the Output:Table:SummaryReports object.

        Args:
            summary (str): Choices are AllSummary, AllMonthly,
                AllSummaryAndMonthly, AllSummaryAndSizingPeriod,
                AllSummaryMonthlyAndSizingPeriod,
                AnnualBuildingUtilityPerformanceSummary,
                InputVerificationandResultsSummary,
                SourceEnergyEndUseComponentsSummary, ClimaticDataSummary,
                EnvelopeSummary, SurfaceShadowingSummary, ShadingSummary,
                LightingSummary, EquipmentSummary, HVACSizingSummary,
                ComponentSizingSummary, CoilSizingDetails, OutdoorAirSummary,
                SystemSummary, AdaptiveComfortSummary, SensibleHeatGainSummary,
                Standard62.1Summary, EnergyMeters, InitializationSummary,
                LEEDSummary, TariffReport, EconomicResultSummary,
                ComponentCostEconomicsSummary, LifeCycleCostReport,
                HeatEmissionsSummary,
        Returns:
            Outputs: self
        """
        outputs = [
            {
                "key": "Output:Table:SummaryReports".upper(),
                **dict(Report_1_Name=summary),
            }
        ]

        self._outputs.extend(outputs)
        return self

    def add_sql(self, sql_output_style="SimpleAndTabular"):
        """Adds the `Output:SQLite` object. This object will produce an sql file
        that contains the simulation results in a database format. See
        `eplusout.sql
        <https://bigladdersoftware.com/epx/docs/9-2/output-details-and
        -examples/eplusout-sql.html#eplusout.sql>`_ for more details.

        Args:
            sql_output_style (str): The *Simple* option will include all of the
                predefined database tables as well as time series related data.
                Using the *SimpleAndTabular* choice adds database tables related
                to the tabular reports that are already output by EnergyPlus in
                other formats.
        Returns:
            Outputs: self
        """
        output = {"key": "Output:SQLite".upper(), **dict(Option_Type=sql_output_style)}

        self._outputs.extend([output])
        return self

    def add_output_control(self, output_control_table_style="CommaAndHTML"):
        """Sets the `OutputControl:Table:Style` object.

        Args:
            output_control_table_style (str): Choices are: Comma, Tab, Fixed,
                HTML, XML, CommaAndHTML, TabAndHTML, XMLAndHTML, All
        Returns:
            Outputs: self
        """
        outputs = [
            {
                "key": "OutputControl:Table:Style".upper(),
                **dict(Column_Separator=output_control_table_style),
            }
        ]

        self._outputs.extend(outputs)
        return self

    def add_umi_template_outputs(self):
        """Adds the necessary outputs in order to create an UMI template."""
        # list the outputs here
        outputs = [
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Air System Total Heating Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Air System Total Cooling Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Zone Ideal Loads Zone Total Cooling Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Zone Ideal Loads Zone Total Heating Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Zone Thermostat Heating Setpoint Temperature",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Zone Thermostat Cooling Setpoint Temperature",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Heat Exchanger Total Heating Rate",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Heat Exchanger Sensible Effectiveness",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Heat Exchanger Latent Effectiveness",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Water Heater Heating Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Air System Outdoor Air Minimum Flow Fraction",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "OUTPUT:METER",
                **dict(
                    Key_Name="HeatRejection:EnergyTransfer",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Heating:EnergyTransfer", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Cooling:EnergyTransfer", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(
                    Key_Name="Heating:DistrictHeating", Reporting_Frequency="hourly"
                ),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Heating:Electricity", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Heating:Gas", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(
                    Key_Name="Cooling:DistrictCooling", Reporting_Frequency="hourly"
                ),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Cooling:Electricity", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Cooling:Electricity", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Cooling:Gas", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(
                    Key_Name="WaterSystems:EnergyTransfer", Reporting_Frequency="hourly"
                ),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Fans:Electricity", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Pumps:Electricity", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(
                    Key_Name="Refrigeration:Electricity", Reporting_Frequency="hourly"
                ),
            },
            {
                "key": "OUTPUT:METER",
                **dict(
                    Key_Name="Refrigeration:EnergyTransfer",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Meter".upper(),
                **dict(
                    Key_Name="HeatingCoils:EnergyTransfer",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Meter".upper(),
                **dict(
                    Key_Name="Baseboard:EnergyTransfer",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Meter".upper(),
                **dict(
                    Key_Name="HeatRejection:Electricity",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Meter".upper(),
                **dict(
                    Key_Name="CoolingCoils:EnergyTransfer",
                    Reporting_Frequency="hourly",
                ),
            },
        ]

        self._outputs.extend(outputs)
        return self

    def add_dxf(self):
        outputs = [
            {
                "key": "Output:Surfaces:Drawing".upper(),
                **dict(Report_Type="DXF", Report_Specifications_1="ThickPolyline"),
            }
        ]
        self._outputs.extend(outputs)
        return self

    def add_umi_ouputs(self):
        """Adds the necessary outputs in order to return the same energy profile
        as in UMI.
        """
        # list the outputs here
        outputs = [
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Air System Total Heating Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Air System Total Cooling Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Zone Ideal Loads Zone Total Cooling Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Zone Ideal Loads Zone Total Heating Energy",
                    Reporting_Frequency="hourly",
                ),
            },
            {
                "key": "Output:Variable".upper(),
                **dict(
                    Variable_Name="Water Heater Heating Energy",
                    Reporting_Frequency="hourly",
                ),
            },
        ]

        self._outputs.extend(outputs)
        return self

    def add_profile_gas_elect_ouputs(self):
        """Adds the following meters: Electricity:Facility, Gas:Facility,
        WaterSystems:Electricity, Heating:Electricity, Cooling:Electricity
        """
        # list the outputs here
        outputs = [
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Electricity:Facility", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Gas:Facility", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(
                    Key_Name="WaterSystems:Electricity", Reporting_Frequency="hourly"
                ),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Heating:Electricity", Reporting_Frequency="hourly"),
            },
            {
                "key": "OUTPUT:METER",
                **dict(Key_Name="Cooling:Electricity", Reporting_Frequency="hourly"),
            },
        ]
        self._outputs.extend(outputs)
        return self

    def apply(self):
        """Applies the outputs to the idf model. Modifies the model by calling
        :meth:`~archetypal.idfclass.idf.IDF.newidfobject`"""
        for output in self._outputs:
            self.idf.newidfobject(**output)
        return self



"""archetypal StructureInformation."""

import collections

from validator_collection import validators

from archetypal.template.constructions.base_construction import ConstructionBase
from archetypal.template.materials.opaque_material import OpaqueMaterial


class MassRatio(object):
    """Handles the properties of the mass ratio for building template structure."""

    __slots__ = ("_high_load_ratio", "_material", "_normal_ratio")

    def __init__(self, HighLoadRatio=None, Material=None, NormalRatio=None, **kwargs):
        """Initialize a MassRatio object with parameters.

        Args:
            HighLoadRatio (float):
            Material (OpaqueMaterial):
            NormalRatio (float):
        """
        self.HighLoadRatio = HighLoadRatio
        self.Material = Material
        self.NormalRatio = NormalRatio

    @property
    def HighLoadRatio(self):
        """Get or set the high load ratio [kg/m2]."""
        return self._high_load_ratio

    @HighLoadRatio.setter
    def HighLoadRatio(self, value):
        self._high_load_ratio = validators.float(value, minimum=0)

    @property
    def Material(self):
        """Get or set the structure OpaqueMaterial."""
        return self._material

    @Material.setter
    def Material(self, value):
        assert isinstance(
            value, OpaqueMaterial
        ), f"Material must be of type OpaqueMaterial, not {type(value)}"
        self._material = value

    @property
    def NormalRatio(self):
        """Get or set the normal load ratio [kg/m2]."""
        return self._normal_ratio

    @NormalRatio.setter
    def NormalRatio(self, value):
        self._normal_ratio = validators.float(value, minimum=0)

    def __hash__(self):
        """Return the hash value of self."""
        return hash(self.__key__())

    def __key__(self):
        """Get a tuple of attributes. Useful for hashing and comparing."""
        return (
            self.HighLoadRatio,
            self.Material,
            self.NormalRatio,
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, MassRatio):
            return NotImplemented
        else:
            return self.__key__() == other.__key__()

    def __iter__(self):
        """Iterate over attributes. Yields tuple of (keys, value)."""
        for k, v in self.mapping().items():
            yield k, v

    def to_dict(self):
        """Return MassRatio dictionary representation."""
        return collections.OrderedDict(
            HighLoadRatio=self.HighLoadRatio,
            Material={"$ref": str(self.Material.id)},
            NormalRatio=self.NormalRatio,
        )

    def mapping(self):
        """Get a dict based on the object properties, useful for dict repr."""
        return dict(
            HighLoadRatio=self.HighLoadRatio,
            Material=self.Material,
            NormalRatio=self.NormalRatio,
        )

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return self

    @classmethod
    def generic(cls):
        """Create generic MassRatio object."""
        mat = OpaqueMaterial(
            Name="Steel General",
            Conductivity=45.3,
            SpecificHeat=500,
            SolarAbsorptance=0.4,
            ThermalEmittance=0.9,
            VisibleAbsorptance=0.4,
            Roughness="Rough",
            Cost=0,
            Density=7830,
            MoistureDiffusionResistance=50,
            EmbodiedCarbon=1.37,
            EmbodiedEnergy=20.1,
            TransportCarbon=0.067,
            TransportDistance=500,
            TransportEnergy=0.94,
            SubstitutionRatePattern=[1],
            SubstitutionTimestep=100,
            DataSource="BostonTemplateLibrary.json",
        )
        return cls(HighLoadRatio=305, Material=mat, NormalRatio=305)

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(self.HighLoadRatio, self.Material, self.NormalRatio)


class StructureInformation(ConstructionBase):
    """Building Structure settings.

    .. image:: ../images/template/constructions-structure.png
    """

    __slots__ = ("_mass_ratios",)

    def __init__(self, Name, MassRatios, **kwargs):
        """Initialize object.

        Args:
            MassRatios (list of MassRatio): MassRatio object.
            **kwargs: keywords passed to the ConstructionBase constructor.
        """
        super(StructureInformation, self).__init__(Name, **kwargs)
        self.MassRatios = MassRatios

    @property
    def MassRatios(self):
        """Get or set the list of MassRatios."""
        return self._mass_ratios

    @MassRatios.setter
    def MassRatios(self, value):
        assert isinstance(value, list), "mass_ratio must be of a list of MassRatio"
        self._mass_ratios = value

    @classmethod
    def from_dict(cls, data, materials, **kwargs):
        """Create StructureInformation from a dictionary.

        Args:
            data (dict): A python dictionary.
            materials (dict): A dictionary of python OpaqueMaterials with their id as
                keys.
            **kwargs: keywords passed to parent constructors.
        """
        mass_ratio_ref = data.pop("MassRatios")
        mass_ratios = [
            MassRatio(
                HighLoadRatio=massratio["HighLoadRatio"],
                Material=materials[massratio["Material"]["$ref"]],
                NormalRatio=massratio["NormalRatio"],
            )
            for massratio in mass_ratio_ref
        ]
        _id = data.pop("$id")
        return cls(MassRatios=mass_ratios, id=_id, **data, **kwargs)

    def to_dict(self):
        """Return StructureInformation dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["MassRatios"] = [mass.to_dict() for mass in self.MassRatios]
        data_dict["AssemblyCarbon"] = self.AssemblyCarbon
        data_dict["AssemblyCost"] = self.AssemblyCost
        data_dict["AssemblyEnergy"] = self.AssemblyEnergy
        data_dict["DisassemblyCarbon"] = self.DisassemblyCarbon
        data_dict["DisassemblyEnergy"] = self.DisassemblyEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def validate(self):
        """Validate object and fill in missing values."""
        return self

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            MassRatios=self.MassRatios,
            AssemblyCarbon=self.AssemblyCarbon,
            AssemblyCost=self.AssemblyCost,
            AssemblyEnergy=self.AssemblyEnergy,
            DisassemblyCarbon=self.DisassemblyCarbon,
            DisassemblyEnergy=self.DisassemblyEnergy,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __hash__(self):
        """Return the hash value of self."""
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, StructureInformation):
            return NotImplemented
        else:
            return all(
                [
                    self.AssemblyCarbon == other.AssemblyCarbon,
                    self.AssemblyCost == other.AssemblyCost,
                    self.AssemblyEnergy == other.AssemblyEnergy,
                    self.DisassemblyCarbon == other.DisassemblyCarbon,
                    self.DisassemblyEnergy == other.DisassemblyEnergy,
                    self.MassRatios == other.MassRatios,
                ]
            )

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(**self.mapping(validate=False))



"""archetypal MaterialLayer."""

import collections
import logging as lg

from sigfig import round
from validator_collection import validators

from archetypal.utils import log


class MaterialLayer(object):
    """Class used to define one layer in a construction assembly.

    This class has two attributes:

    1. Material (:class:`OpaqueMaterial` or :class:`GlazingMaterial` or
       :class:`GasMaterial`): the material object for this layer.
    2. Thickness (float): The thickness of the material in the layer.
    """

    __slots__ = ("_material", "_thickness")

    def __init__(self, Material, Thickness, **kwargs):
        """Initialize a MaterialLayer object with parameters.

        Args:
            Material (OpaqueMaterial, GlazingMaterial, GasMaterial):
            Thickness (float): The thickness of the material in the
                construction.
        """
        self.Material = Material
        self.Thickness = Thickness

    @property
    def Material(self):
        """Get or set the material of self."""
        return self._material

    @Material.setter
    def Material(self, value):
        from archetypal.template.materials import GlazingMaterial, OpaqueMaterial

        assert isinstance(value, (OpaqueMaterial, GlazingMaterial)), (
            f"Input value error for '{value}'. Value must be "
            f"of type (OpaqueMaterial, GlazingMaterial), not {type(value)}"
        )
        self._material = value

    @property
    def Thickness(self):
        """Get or set the material thickness [m]."""
        return self._thickness

    @Thickness.setter
    def Thickness(self, value):
        self._thickness = value
        if value < 0.003:
            log(
                "Modeling layer thinner (less) than 0.003 m (not recommended) for "
                f"MaterialLayer '{self}'",
                lg.WARNING,
            )

    @property
    def resistivity(self):
        """Get or set the resistivity of the material layer [m-K/W]."""
        return 1 / self.Material.Conductivity

    @resistivity.setter
    def resistivity(self, value):
        self.Material.Conductivity = 1 / validators.float(value, minimum=0)

    @property
    def r_value(self):
        """Get or set the the R-value of the material layer [m2-K/W].

        Note that, when setting the R-value, the thickness of the material will
        be adjusted and the conductivity will remain fixed.
        """
        return self.Thickness / self.Material.Conductivity

    @r_value.setter
    def r_value(self, value):
        self.Thickness = validators.float(value, minimum=0) * self.Material.Conductivity

    @property
    def u_value(self):
        """Get or set the heat transfer coefficient [W/(m2⋅K)]."""
        return 1 / self.r_value

    @u_value.setter
    def u_value(self, value):
        self.r_value = 1 / validators.float(value, minimum=0)

    @property
    def heat_capacity(self):
        """Get the material layer's heat capacity [J/(m2-k)]."""
        return self.Material.Density * self.Material.SpecificHeat * self.Thickness

    @property
    def specific_heat(self):
        """Get the material layer's specific heat [J/kg-K]."""
        return self.Material.SpecificHeat

    def to_dict(self):
        """Return MaterialLayer dictionary representation."""
        return collections.OrderedDict(
            Material={"$ref": str(self.Material.id)},
            Thickness=round(self.Thickness, decimals=3),
        )

    def to_epbunch(self, idf):
        """Convert self to an EpBunch given an IDF model.

        Notes:
            The object is added to the idf model.
            The thickness is passed to the epbunch.

        Args:
            idf (IDF): An IDF model.

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return self.Material.to_epbunch(idf, self.Thickness)

    def mapping(self):
        """Get a dict based on the object properties, useful for dict repr."""
        return dict(Material=self.Material, Thickness=self.Thickness)

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return self

    def __hash__(self):
        """Return the hash value of self."""
        return id(self)

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, MaterialLayer):
            return NotImplemented
        else:
            return all(
                [self.Thickness == other.Thickness, self.Material == other.Material]
            )

    def __repr__(self):
        """Return a representation of self."""
        return "{} with thickness of {:,.3f} m".format(self.Material, self.Thickness)

    def __iter__(self):
        """Iterate over attributes. Yields tuple of (keys, value)."""
        for k, v in self.mapping().items():
            yield k, v

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(self.Material, self.Thickness)



import pytest

from archetypal import IDF
from archetypal.zone_graph import ZoneGraph


class TestZoneGraph:
    """Series of tests for the :class:`ZoneGraph` class"""

    @pytest.fixture(scope="class")
    def small_office(config):
        file = (
            "tests/input_data/necb/NECB 2011-SmallOffice-NECB HDD "
            "Method-CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw.idf"
        )
        w = "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw"
        idf = IDF(file, epw=w)
        yield idf

    def test_traverse_graph(self, small_office):
        """
        Args:
            small_office:
        """

        idf = small_office

        G = ZoneGraph.from_idf(idf, log_adj_report=False)

        assert G

    @pytest.fixture(scope="class")
    def G(self, config, small_office):
        """
        Args:
            config:
            small_office:
        """

        idf = small_office
        yield ZoneGraph.from_idf(idf)

    @pytest.mark.parametrize("adj_report", [True, False])
    def test_graph(self, small_office, adj_report):
        """Test the creation of a BuildingTemplate zone graph. Parametrize the
        creation of the adjacency report

        Args:
            small_office:
            adj_report:
        """
        import networkx as nx

        idf = small_office

        G1 = ZoneGraph.from_idf(idf, log_adj_report=adj_report)
        assert not nx.is_empty(G1)
        from eppy.bunch_subclass import EpBunch

        assert isinstance(
            G1.nodes["Sp-Attic Sys-0 Flr-2 Sch-- undefined - HPlcmt-core ZN"][
                "epbunch"
            ],
            EpBunch,
        )

    def test_graph_info(self, G):
        """test the info method on a ZoneGraph

        Args:
            G:
        """
        G.info()

    def test_viewgraph2d(self, G):
        """test the visualization of the zonegraph in 2d

        Args:
            G:
        """
        import networkx as nx

        G.plot_graph2d(
            nx.layout.circular_layout,
            (1),
            font_color="w",
            legend=True,
            font_size=8,
            color_nodes="core",
            node_labels_to_integers=True,
            plt_style="seaborn",
            save=False,
            show=False,
            filename="test",
        )

    @pytest.mark.parametrize("annotate", [True, "Name", ("core", None)])
    def test_viewgraph3d(self, G, annotate):
        """test the visualization of the zonegraph in 3d

        Args:
            G:
            annotate:
        """
        G.plot_graph3d(
            annotate=annotate,
            axis_off=True,
            save=False,
            show=False,
        )

    def test_core_graph(self, G):
        """
        Args:
            G:
        """
        H = G.core_graph

        assert len(H) == 1  # assert G has no nodes since Warehouse does not have a
        # core zone

    def test_perim_graph(self, G):
        """
        Args:
            G:
        """
        H = G.perim_graph

        assert len(H) > 0  # assert G has at least one node



import os

import pytest
from click.testing import CliRunner
from path import Path

from archetypal import settings
from archetypal.cli import cli
from archetypal.utils import log


class TestCli:
    """Defines tests for usage of the archetypal Command Line Interface"""

    def test_reduce(self):
        """Tests the 'reduced_model' method"""
        runner = CliRunner()
        base = Path("tests/input_data/umi_samples")
        outname = "tests/.temp/warehouse.json"
        result = runner.invoke(
            cli,
            [
                "-csd",
                "--cache-folder",
                "tests/.temp/cache",
                "--data-folder",
                "tests/.temp/data",
                "--imgs-folder",
                "tests/.temp/images",
                "--logs-folder",
                "tests/.temp/logs",
                "--ep_version",
                settings.ep_version,
                "reduce",
                "-w",
                "tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_*.epw",
                base / "*Res*.idf",
                "-o",
                outname,
            ],
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        assert Path(outname).exists()

    @pytest.mark.skipif(
        os.environ.get("CI", "False").lower() == "true",
        reason="Skipping this test on CI environment.",
    )
    def test_transition_dir_file_mixed(self):
        """Tests the transition method for the CLI using a mixture of a directory
        (Path.isdir()) and a file Path.isfile()"""
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--cache-folder",
                "tests/.temp/cache",
                "--data-folder",
                "tests/.temp/data",
                "--imgs-folder",
                "tests/.temp/images",
                "--logs-folder",
                "tests/.temp/logs",
                "transition",
                "-v",
                "9.2",
                "tests/input_data/problematic/ASHRAE90.1_ApartmentHighRise_STD2016_Buffalo.idf",
                "tests/input_data/problematic/*.idf",  # Path with wildcard
                "tests/input_data/problematic",  # Just a path
            ],
            catch_exceptions=False,
        )
        log(result.stdout)
        assert result.exit_code == 0
